"""TODO(MyDataset): Add a description here."""

import json
import os

import datasets

_DESCRIPTION = """\
 Mol-Instructions datasets.
"""
_DATA_URL = "/home/hukun/code/trimodal_code/craft_mol/data/Mol-Instructions/data"


class DatasetConfig(datasets.BuilderConfig):
    def __init__(self, data_url, **kwargs):
        """BuilderConfig for MyDataset

        Args:
          data_url: `string`, url to the dataset (word or raw level)
          **kwargs: keyword arguments forwarded to super.
        """
        super(DatasetConfig, self).__init__(
            version=datasets.Version(
                "1.0.0",
            ),
            **kwargs,
        )
        self.data_url = data_url


class MyDataset(datasets.GeneratorBasedBuilder):
    VERSION = datasets.Version("0.1.0")
    BUILDER_CONFIGS = [
        DatasetConfig(
            name="Molecule-oriented Instructions",
            data_url=_DATA_URL + "/" + "Molecule-oriented_Instructions.zip",
            description="Molecule-oriented Instructions",
        ),
        DatasetConfig(
            name="Protein-oriented Instructions",
            data_url=_DATA_URL + "/" + "Protein-oriented_Instructions.zip",
            description="Protein-oriented Instructions",
        ),
        DatasetConfig(
            name="Biomolecular Text Instructions",
            data_url=_DATA_URL + "/" + "Biomolecular_Text_Instructions.zip",
            description="Biomolecular Text Instructions",
        ),
    ]

    def _info(self):
        return datasets.DatasetInfo(
            # This is the description that will appear on the datasets page.
            description=_DESCRIPTION,
            features=datasets.Features(
                {
                    "instruction": datasets.Value("string"),
                    "input": datasets.Value("string"),
                    "output": datasets.Value("string"),
                    "metadata": datasets.Value("string"),
                    # These are the features of your dataset like images, labels ...
                }
            ),
            # If there's a common (input, target) tuple from the features,
            # specify them here. They'll be used if as_supervised=True in
            # builder.as_dataset.
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        if self.config.name == "Molecule-oriented Instructions":
            data_file = dl_manager.download_and_extract(self.config.data_url)
            data_dir = os.path.join(data_file, "Molecule-oriented_Instructions")
            
            return [
                datasets.SplitGenerator(
                    name="description_guided_molecule_design",
                    gen_kwargs={"filepath": os.path.join(data_dir, "description_guided_molecule_design.json")},
                ),
                datasets.SplitGenerator(
                    name="forward_reaction_prediction",
                    gen_kwargs={"filepath": os.path.join(data_dir, "forward_reaction_prediction.json")},
                ),
                datasets.SplitGenerator(
                    name="molecular_description_generation",
                    gen_kwargs={"filepath": os.path.join(data_dir, "molecular_description_generation.json")},
                ),
                datasets.SplitGenerator(
                    name="property_prediction",
                    gen_kwargs={"filepath": os.path.join(data_dir, "property_prediction.json")},
                ),
                datasets.SplitGenerator(
                    name="reagent_prediction",
                    gen_kwargs={"filepath": os.path.join(data_dir, "reagent_prediction.json")},
                ),
                datasets.SplitGenerator(
                    name="retrosynthesis",
                    gen_kwargs={"filepath": os.path.join(data_dir, "retrosynthesis.json")},
                ),
            ]
        elif self.config.name == "Protein-oriented Instructions":
            data_file = dl_manager.download_and_extract(self.config.data_url)
            data_dir = os.path.join(data_file, "Protein-oriented_Instructions")
            
            return [
                datasets.SplitGenerator(
                    name="catalytic_activity",
                    gen_kwargs={"filepath": os.path.join(data_dir, "catalytic_activity.json")},
                ),
                datasets.SplitGenerator(
                    name="domain_motif",
                    gen_kwargs={"filepath": os.path.join(data_dir, "domain_motif.json")},
                ),
                datasets.SplitGenerator(
                    name="protein_function",
                    gen_kwargs={"filepath": os.path.join(data_dir, "protein_function.json")},
                ),
                datasets.SplitGenerator(
                    name="general_function",
                    gen_kwargs={"filepath": os.path.join(data_dir, "general_function.json")},
                ),
                datasets.SplitGenerator(
                    name="protein_design",
                    gen_kwargs={"filepath": os.path.join(data_dir, "protein_design.json")},
                ),
            ]
        elif self.config.name == "Biomolecular Text Instructions":
            data_file = dl_manager.download_and_extract(self.config.data_url)
            data_dir = os.path.join(data_file, "Biomolecular_Text_Instructions")
            
            return [
                datasets.SplitGenerator(
                    # name=datasets.Split.ALL,
                    name="chemical_disease_interaction_extraction",
                    gen_kwargs={"filepath": os.path.join(data_dir, "chemical_disease_interaction_extraction.json")},
                ),
                datasets.SplitGenerator(
                    name="chemical_entity_recognition",
                    gen_kwargs={"filepath": os.path.join(data_dir, "chemical_entity_recognition.json")},
                ),
                datasets.SplitGenerator(
                    name="chemical_protein_interaction_extraction",
                    gen_kwargs={"filepath": os.path.join(data_dir, "chemical_protein_interaction_extraction.json")},
                ),
                datasets.SplitGenerator(
                    name="multi_choice_question",
                    gen_kwargs={"filepath": os.path.join(data_dir, "multi_choice_question.json")},
                ),
                datasets.SplitGenerator(
                    name="open_question",
                    gen_kwargs={"filepath": os.path.join(data_dir, "open_question.json")},
                ),
                datasets.SplitGenerator(
                    name="true_or_false_question",
                    gen_kwargs={"filepath": os.path.join(data_dir, "true_or_false_question.json")},
                ),
            ]

    def _generate_examples(self, filepath):
        """Yields examples."""
        # TODO(sciQ): Yields (key, example) tuples from the dataset
        with open(filepath, encoding="utf-8") as f:
            data = json.load(f)
            for id_, row in enumerate(data):
                yield id_, row





import torch
from torch import nn
from model.TMMF import TMMF
from dataset import create_dataset,create_dataloader
import ruamel.yaml as yaml
import torch.nn.functional as F
import numpy as np
import argparse
from pathlib import Path
from rdkit import Chem
from rdkit.Chem import Draw
from dataset.tokenizer import Iupac_tokenizer,Selfies_tokenizer
import selfies as sf
import os
import plotly.graph_objects as go
import matplotlib.pyplot as plt


class Config(dict):
    def __init__(self, *args, **kwargs):
        super(Config, self).__init__(*args, **kwargs)

    def __getattr__(self, key):
        value = self[key]
        if isinstance(value, dict):
            value = Config(value)
        return value

class BertForInteract(nn.Module):

    def __init__(self,selfies_encoder,iupac_encoder,mol_encoder,fusion_encoder):
        super().__init__()
        self.selfies_encoder = selfies_encoder
        self.iupac_encoder = iupac_encoder
        self.mol_encoder = mol_encoder
        self.fusion_encoder = fusion_encoder
        for p in self.parameters():
            p.requires_grad=False

    def forward(self,selfies_ids=None,iupac_ids=None,mol_ids=None,labels=None,save_attention=True):
        selfies_embeds,selfies_mask = self.selfies_encoder(input_ids=selfies_ids,mode='selfies')
        iupac_embeds,iupac_mask = self.iupac_encoder(input_ids=iupac_ids,mode='iupac')
        mol_pool,mol_embeds,mol_mask = self.mol_encoder(mol_ids)



        iupac_mol_embeds = torch.cat([iupac_embeds,mol_embeds],dim=1)
        iupac_mol_mask = torch.cat([iupac_mask,mol_mask],dim=1)
        output = self.fusion_encoder(selfies_embeds=selfies_embeds,
                                     iupac_mol_embeds=iupac_mol_embeds,
                                     selfies_mask=selfies_mask,
                                     iupac_mol_mask=iupac_mol_mask,
                                     mode='fusion',save_attention=save_attention)

        return output
    
def find_satisfy_iupac(dataset,iupac_token,iupac_tokenizer,sample_num=10):
    ids_sat = []        # save ids satisfied
    position = [] # record iupac_token position
    for i in range(len(dataset)):
        token_position = []
        iupac = dataset[i][2][0]['iupac_tokens']
        tokens = iupac_tokenizer.convert_ids_to_tokens(iupac.tolist())
        length = len(iupac) - iupac.tolist().count(0)
        if length < 40:
        # iupac token是一个list
            for j in range(len(iupac_token)):
                if iupac_token[j] in tokens:
                    token_position.append(tokens.index(iupac_token[j]))
            if len(token_position) == len(iupac_token):
                ids_sat.append(i)
                position.append(token_position)

            if len(ids_sat) == sample_num:
                break
    return ids_sat,position

def find_satisfy_iupac_cer(dataset,iupac_token,iupac_tokenizer,sample_num=10):
    ids_sat = []        # save ids satisfied
    position = [] # record iupac_token position
    for i in range(len(dataset)):
        token_position = []
        iupac = dataset[i][2][0]['iupac_tokens']
        iupac_token_list = iupac.tolist()
        iupac_token_list = [i for i in iupac_token_list if i!=0 and i!=1 and i!=2]
        tokens = iupac_tokenizer.convert_ids_to_tokens(iupac_token_list)
        string = ''.join(tokens)
        for j in range(len(iupac_token)):
            if string == iupac_token[j]:
                ids_sat.append(i)
                print(string)
        if len(ids_sat) == len(iupac_token):
            break
    return ids_sat,position


def get_heatmap_iupac_selfies(idx,ids_sat,token_position,dataset,attention,selfies_tokenizer,iupac_tokenizer):  
    index = ids_sat[idx]
    selfies_sample = dataset[index][1][0]['selfies_tokens']
    iupac_sample = dataset[index][2][0]['iupac_tokens']
    selfies_tokens = selfies_tokenizer.convert_ids_to_tokens(selfies_sample.tolist())   

    pos = token_position[idx]   # pos is a list
    iupac_tokens = iupac_tokenizer.convert_ids_to_tokens(iupac_sample.tolist())

    heatmap_name = []
    selfies_tokens_valid = []
    iupac_tokens_valid = []
    for j in range(1,attention[0].shape[0]):
        if selfies_tokens[j] == '[PAD]':
            break
        selfies_tokens_valid.append(selfies_tokens[j])
    try:
        smi = sf.decoder("".join(selfies_tokens_valid[0:-1]))
    except:
        smi = None
    selfies_tokens_valid.pop(-1)
    for po in pos:
        iupac_tokens_valid.append(iupac_tokens[po])
    # iupac_tokens_valid.append(iupac_tokens[pos])
    # print(selfies_tokens_valid)
    # print(iupac_tokens_valid)

    for k in range(attention.shape[0]): 
        attention_head = attention[k]
        attention_name = attention_head[1:j-1,pos]
        heatmap_name.append(attention_name.cpu())

    heatmap_name = np.array(heatmap_name)

    return heatmap_name,selfies_tokens_valid,iupac_tokens_valid,smi

def statistics_name_topk(heatmap_name,selfies_tokens_valid,iupac_tokens_valid,topk=5):
    attn_matrix = heatmap_name[-1]
    attn_matrix = np.array(attn_matrix)

    attn_rank = attn_matrix.copy()

    topk_selfies_tokens = []
    topk_coefficient = []

    # find topk selfies_tokens and record
    for i in range(topk):
        max_index = np.argmax(attn_rank)
        topk_selfies_tokens.append(selfies_tokens_valid[max_index])
        topk_coefficient.append(attn_matrix[max_index][0])
        attn_rank[max_index] = 0

    return topk_selfies_tokens,topk_coefficient
    
def statistics_atom_topk(heatmap_atom,selfies_tokens_valid,atom_tokens_valid,topk=5):
    attn_matrix = heatmap_atom[-1]
    attn_matrix = np.array(attn_matrix)
    attn_matrix = attn_matrix.T

    attn_rank = attn_matrix.copy()

    topk_atom_tokens = []
    topk_coefficient = []

    # find topk selfies_tokens and record
    for i in range(topk):
        max_index = np.argmax(attn_rank)
        topk_atom_tokens.append(atom_tokens_valid[max_index])
        topk_coefficient.append(attn_matrix[max_index][0])
        attn_rank[max_index] = 0

    return topk_atom_tokens,topk_coefficient


def plot_heatmap_name(heatmap_name,selfies_tokens_valid,iupac_tokens_valid):
    import matplotlib.pyplot as plt
    import seaborn as sns
    import matplotlib.gridspec as gridspec
    # heatmap_name shape (num_heatmaps, rows, columns)


    # num_heatmaps = heatmap_name.shape[0]

    import numpy as np

    # 注意力系数矩阵
    attn_matrix = heatmap_name[-1]
    attn_matrix = np.array(attn_matrix)


    num_selfies_tokens = len(selfies_tokens_valid)
    num_iupac_tokens = len(iupac_tokens_valid)


    # 定义数据
    source_tokens = iupac_tokens_valid
    target_tokens = selfies_tokens_valid
    correlation_matrix = attn_matrix.T

    # 自动生成颜色
    cmap = plt.get_cmap('Pastel1')
    colors = [cmap(i) for i in range(len(source_tokens))]
    colors = ['rgba({},{},{},{})'.format(int(r*255), int(g*255), int(b*255), a) for r, g, b, a in colors]

    # 构建Sankey图数据结构
    source = []
    target = []
    value = []
    link_colors = []

    node_colors = colors + ['#E0E0E0'] * len(target_tokens) 

    for i, src_token in enumerate(source_tokens):
        for j, tgt_token in enumerate(target_tokens):
            source.append(i)
            target.append(j + len(source_tokens))
            value.append(correlation_matrix[i][j])
            link_colors.append(colors[i])  

    # # 绘制Sankey图
    # fig1 = go.Figure(data=[go.Sankey(
    #     node=dict(
    #         pad=25,
    #         thickness=30,  # 调整节点宽度
    #         line=dict(color="black", width=0.5),
    #         label=[],
    #         color=node_colors,
    #     ),
    #     link=dict(
    #         source=source,
    #         target=target,
    #         value=value,
    #         color=link_colors
    #     )
    # )])
    fig2 = go.Figure(data=[go.Sankey(
    node=dict(
        pad=25,
        thickness=30,  # 调整节点宽度
        line=dict(color="black", width=0.5),
        label=source_tokens + target_tokens,
        color=node_colors,
    ),
    link=dict(
        source=source,
        target=target,
        value=value,
        color=link_colors
            )
        )])

    # fig1.update_layout(title_text="Sankey Diagram of Token Correlation ", font_size=10)
    # fig2.update_layout(title_text="Sankey Diagram of Token Correlation ", font_size=10)
    # 创建一个新的图
    # fig = plt.figure(figsize=(10, 15))
    # gs = gridspec.GridSpec(1, 2, width_ratios=[0.5,0.01])

    # ax = plt.subplot(gs[0, 0])
    # # 计算每个token的位置
    # positions_selfies = np.array([(x, 0) for x in range(num_selfies_tokens)])
    # positions_iupac = np.array([(x+num_selfies_tokens/2, 1) for x in range(num_iupac_tokens)])
    # # print(positions_selfies)
    # # print(positions_iupac)

    # # 画方块和token文字
    # for i, pos in enumerate(np.concatenate((positions_selfies, positions_iupac))):
    #     color = (255/255,223/255,146/255) if i < num_selfies_tokens else '#f0a0a0'  # 使用饱和度较低的颜色
    #     token = selfies_tokens_valid[i] if i < num_selfies_tokens else iupac_tokens_valid[i - num_selfies_tokens]
    #     if i < num_selfies_tokens:
    #         ax.add_patch(plt.Rectangle((pos[1] - 2, pos[0]), 0.5, 0.8, fill=True, color=color))  # 方块尺寸放大
    #         ax.text(pos[1]-1.75, pos[0]+0.4, token, ha='center', va='center',fontdict={'fontsize': 9, 'fontweight': 'bold'})
    #     else:
    #         ax.add_patch(plt.Rectangle((pos[1], pos[0]+2*(i-num_selfies_tokens)), 0.5, 0.8, fill=True, color=color))  # 方块尺寸放大
    #         ax.text(pos[1]+0.25, pos[0]+2*(i-num_selfies_tokens)+0.4, token, ha='center', va='center',fontdict={'fontsize': 10, 'fontweight': 'bold'})


        

    # # 画线
    # for i in range(num_selfies_tokens):
    #     for j in range(num_iupac_tokens):
    #         if attn_matrix[i, j] > 0.05:  # 注意力系数大于0的连接才画线
    #             # 通过注意力系数来调整线的颜色
    #             color = plt.cm.Blues(attn_matrix[i, j])  # 你可以选择你喜欢的颜色映射
    #             ax.plot(
    #                 [positions_selfies[i, 1]-1.5, positions_iupac[j, 1]],
    #                     [positions_selfies[i, 0]+0.5, positions_iupac[j, 0]+2*j+0.5],
    #                     color=color, linewidth=2.5)  # 这里固定了线宽为2，也可以通过注意力系数来调整

    # # 设置图形的界限和标签
    # ax.set_xlim([-2, 2])
    # ax.set_ylim([-1, max(num_selfies_tokens, num_iupac_tokens)])
    # ax.set_yticks([])  # 这里我们不显示y轴的刻度
    # ax.set_yticklabels([])
    # ax.set_xticks([-1.75, 1.25])
    # ax.set_xticklabels(['SELFIES', 'IUPAC'])
    # # ax.invert_yaxis()  # 反转y轴，使得第一个token位于顶部

    # # 去掉周围的框线
    # for side in ['top', 'right', 'bottom', 'left']:
    #     ax.spines[side].set_visible(False)

    # # 添加颜色条
    # cax = plt.subplot(gs[:, -1])
    # sm = plt.cm.ScalarMappable(cmap=plt.cm.Blues, norm=plt.Normalize(vmin=0, vmax=1))
    # sm.set_array([])
    # cbar = plt.colorbar(sm, ax=ax, label='Attention Score',fraction=0.03,cax=cax)
    
    # plt.show()

    # # rows = int(np.ceil(num_heatmaps / 2))
    # fig = plt.figure(figsize=(15, 10))

    # # 定义一个GridSpec，将画布分为rows行，3列
    # gs = gridspec.GridSpec(8, 2, width_ratios=[0.5,0.01] ,height_ratios=[0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2])


    # for idx in range(num_heatmaps):
    #     ax = plt.subplot(gs[idx % 8, idx // 8])
    #     # 转置heatmap_name[idx]，使得行对应于iupac_tokens_valid，列对应于selfies_tokens_valid
    #     sns.heatmap(heatmap_name[idx].T,ax=ax, cmap='PuBu', cbar=False)
    #     ax.set_title(f'Head {idx + 1}',fontdict={'fontsize': 10, 'fontweight': 'bold','fontfamily':'Operator Mono'})
    #     if (idx+1) % 8 == 0:
    #         ax.set_xticks(np.arange(len(selfies_tokens_valid)) + 0.5, selfies_tokens_valid, rotation=90, fontsize=10)
    #     else:
    #         ax.set_xticks([])
    #     ax.set_yticks(np.arange(len(iupac_tokens_valid)) + 0.5, iupac_tokens_valid, rotation=0, fontsize=10)
    #     # for i in range(len(heatmap_name[idx])):
    #     #     ax.text(0.5,i+0.5,selfies_tokens_valid[i],ha='center',va='center',fontdict={'fontsize': 8, 'fontweight': 'medium'})
    # # cloarbar third column
    # cax = plt.subplot(gs[:, -1])
    # cbar = plt.colorbar(ax.collections[0], cax=cax)
    # cbar.ax.set_xlabel('attention score', rotation=0, labelpad=15)  # add colorbar title

    # plt.tight_layout()
    
    # 将source_tokens target_tokens 以及matrix 保存为csv文件
    import pandas as pd
    df = pd.DataFrame(correlation_matrix,columns=target_tokens,index=source_tokens)
    # 保存为三列，分别是source target value
    # source = []
    # target = []
    # value = []
    # for i, src_token in enumerate(source_tokens):
    #     for j, tgt_token in enumerate(target_tokens):
    #         source.append(src_token)
    #         target.append(f'{j}_'+tgt_token)
    #         value.append(correlation_matrix[i][j])
    # df = pd.DataFrame({'IUPAC':source,'SELFIES':target,'value':value})

    return df,fig2


def find_satify_mol(dataset,selfies_token,selfies_tokenizer,atom_vocab_reverse,sample_num=10):
    ids_sat = []        # save ids satisfied
    position = [] # record iupac_token position
    # find mol contains this iupac_token
    for i in range(len(dataset)):
        token_position = []
        selfies = dataset[i][1][0]['selfies_tokens']
        mol = dataset[i][3]
        tokens = selfies_tokenizer.convert_ids_to_tokens(selfies.tolist())
        try:
            atom_tokens = [atom_vocab_reverse[k] for k in mol.x[:,0].tolist()]
            for j in range(len(selfies_token)):
                if selfies_token[j] in tokens:
                    token_position.append(tokens.index(selfies_token[j]))
            if len(token_position) == len(selfies_token) and len(atom_tokens)<30:
                ids_sat.append(i)
                position.append(token_position)  
            if len(ids_sat) == sample_num:
                break
        except:
            pass
    return ids_sat,position


def get_heatmap_mol_selfies(idx,ids_sat,token_position,dataset,attention,selfies_tokenizer,iupac_tokenizer,atom_vocab_reverse):
    index = ids_sat[idx]
    mol = dataset[index][3]
    selfies = dataset[index][1][0]['selfies_tokens']
    iupac = dataset[index][2][0]['iupac_tokens']
    iupac_tokens = iupac_tokenizer.convert_ids_to_tokens(iupac.tolist())
    selfies_tokens = selfies_tokenizer.convert_ids_to_tokens(selfies.tolist())
    pos = token_position[idx]       #pos is list
    atom_tokens = [atom_vocab_reverse[q] for q in mol.x[:,0].tolist()]
    heatmap_atom = []
    selfies_tokens_valid = []
    atom_tokens_valid = atom_tokens
    for j in range(1,attention[0].shape[0]):
        if selfies_tokens[j] == '[PAD]':
            break
        selfies_tokens_valid.append(selfies_tokens[j])
    smi = sf.decoder("".join(selfies_tokens_valid[0:-1]))
    selfies_tokens_valid.pop(-1)
    selfies_tokens_select = []
    for po in pos:
        selfies_tokens_select.append(selfies_tokens[po])
    # print(selfies_tokens_valid)
    # print(selfies_tokens_select)
    # print(atom_tokens_valid)
    for k in range(attention.shape[0]): 
        attention_head = attention[k]
        attention_atom = attention_head[pos,len(iupac_tokens):len(iupac_tokens)+len(atom_tokens)]
        heatmap_atom.append(attention_atom.cpu())
    
    heatmap_atom = np.array(heatmap_atom)
    return heatmap_atom,selfies_tokens_select,atom_tokens_valid,smi

def plot_heatmap_atom(heatmap_atom,selfies_tokens_select,atom_tokens_valid):
    import matplotlib.pyplot as plt
    import seaborn as sns
    import numpy as np
    import matplotlib.gridspec as gridspec

    # 注意力系数矩阵
    attn_matrix = heatmap_atom[-1]
    attn_matrix = np.array(attn_matrix).T


    num_atom_tokens = len(atom_tokens_valid)
    num_selfies_tokens = len(selfies_tokens_select)

    # 创建一个新的图
    # fig = plt.figure(figsize=(10, 15))
    # gs = gridspec.GridSpec(1, 2, width_ratios=[0.5,0.01])

    # ax = plt.subplot(gs[0, 0])
    # # 计算每个token的位置
    # positions_atom = np.array([(x, 0) for x in range(num_atom_tokens)])
    # positions_selfies = np.array([(x+num_atom_tokens/2, 1) for x in range(num_selfies_tokens)])
    # # print(positions_selfies)
    # # print(positions_iupac)

    # # 画方块和token文字
    # for i, pos in enumerate(np.concatenate((positions_atom, positions_selfies))):
    #     color = (255/255,223/255,146/255) if i < num_atom_tokens else '#f0a0a0'  # 使用饱和度较低的颜色
    #     token = atom_tokens_valid[i] if i < num_atom_tokens else selfies_tokens_select[i - num_atom_tokens]
    #     if i < num_atom_tokens:
    #         ax.add_patch(plt.Rectangle((pos[1] - 2, pos[0]), 0.5, 0.8, fill=True, color=color))  # 方块尺寸放大
    #         ax.text(pos[1]-1.75, pos[0]+0.4, token, ha='center', va='center',fontdict={'fontsize': 9, 'fontweight': 'bold'})
    #     else:
    #         ax.add_patch(plt.Rectangle((pos[1], pos[0]+2*(i-num_atom_tokens)), 0.5, 0.8, fill=True, color=color))  # 方块尺寸放大
    #         ax.text(pos[1]+0.25, pos[0]+2*(i-num_atom_tokens)+0.4, token, ha='center', va='center',fontdict={'fontsize': 10, 'fontweight': 'bold'})


        

    # # 画线
    # for i in range(num_atom_tokens):
    #     for j in range(num_selfies_tokens):
    #         if attn_matrix[i, j] > 0.01:  # 注意力系数大于0的连接才画线
    #             # 通过注意力系数来调整线的颜色
    #             color = plt.cm.Blues(attn_matrix[i, j])  # 你可以选择你喜欢的颜色映射
    #             ax.plot(
    #                 [positions_atom[i, 1]-1.5, positions_selfies[j, 1]],
    #                     [positions_atom[i, 0]+0.5, positions_selfies[j, 0]+2*j+0.5],
    #                     color=color, linewidth=2.5)  # 这里固定了线宽为2，也可以通过注意力系数来调整

    # # 设置图形的界限和标签
    # ax.set_xlim([-2, 2])
    # ax.set_ylim([-1, max(num_selfies_tokens, num_atom_tokens)])
    # ax.set_yticks([])  # 这里我们不显示y轴的刻度
    # ax.set_yticklabels([])
    # ax.set_xticks([-1.75, 1.25])
    # ax.set_xticklabels(['ATOM', 'SELFIES'])
    # # ax.invert_yaxis()  # 反转y轴，使得第一个token位于顶部

    # # 去掉周围的框线
    # for side in ['top', 'right', 'bottom', 'left']:
    #     ax.spines[side].set_visible(False)

    # # 添加颜色条
    # cax = plt.subplot(gs[:, -1])
    # sm = plt.cm.ScalarMappable(cmap=plt.cm.Blues, norm=plt.Normalize(vmin=0, vmax=1))
    # sm.set_array([])
    # cbar = plt.colorbar(sm, ax=ax, label='Attention Score',fraction=0.03,cax=cax)
    # num_heatmaps = heatmap_atom.shape[0]

    # fig = plt.figure(figsize=(10, 10))


    # gs = gridspec.GridSpec(8, 2, width_ratios=[0.5,0.01] ,height_ratios=[0.2,0.2,0.2,0.2,0.2,0.2,0.2,0.2])


    # for idx in range(num_heatmaps):

    #     ax = plt.subplot(gs[idx % 8, idx // 8])

    #     sns.heatmap(heatmap_atom[idx],ax=ax, cmap='PuBu', cbar=False)
    #     ax.set_title(f'Head {idx + 1}',fontdict={'fontsize': 10, 'fontweight': 'bold','fontfamily':'Operator Mono'})
    #     ax.set_yticks(np.arange(len(selfies_tokens_select)) + 0.5, selfies_tokens_select, rotation=0, fontsize=10)
    #     if (idx+1) % 8 == 0:
    #         ax.set_xticks(np.arange(len(atom_tokens_valid)) + 0.5, atom_tokens_valid, rotation=0, fontsize=10)
    #     else:
    #         ax.set_xticks([])

    #     # for i in range(len(atom_tokens_valid)):
    #     #     ax.text(0.5,i+0.5,atom_tokens_valid[i],ha='center',va='center',fontdict={'fontsize': 8, 'fontweight': 'medium'})

    # cax = plt.subplot(gs[:, -1])
    # cbar = plt.colorbar(ax.collections[0], cax=cax)
    # cbar.ax.set_ylabel('attention score', rotation=90, labelpad=15)  

    # plt.tight_layout()

    # 与plot_heatmap_name 相同，计算source target value
    source_tokens = selfies_tokens_select
    target_tokens = atom_tokens_valid
    correlation_matrix = attn_matrix.T

    # 自动生成颜色
    cmap = plt.get_cmap('Pastel1')
    colors = [cmap(i) for i in range(len(source_tokens))]
    colors = ['rgba({},{},{},{})'.format(int(r*255), int(g*255), int(b*255), a) for r, g, b, a in colors]

    # 构建Sankey图数据结构
    source = []
    target = []
    value = []
    link_colors = []

    node_colors = colors + ['#E0E0E0'] * len(target_tokens) 

    for i, src_token in enumerate(source_tokens):
        for j, tgt_token in enumerate(target_tokens):
            source.append(i)
            target.append(j + len(source_tokens))
            value.append(correlation_matrix[i][j])
            link_colors.append(colors[i])  

    fig2 = go.Figure(data=[go.Sankey(
    node=dict(
        pad=25,
        thickness=30,  # 调整节点宽度
        line=dict(color="black", width=0.5),
        label=source_tokens + target_tokens,
        color=node_colors,
    ),
    link=dict(
        source=source,
        target=target,
        value=value,
        color=link_colors
            )
        )])
    
    # 保存为csv
    import pandas as pd
    df = pd.DataFrame(correlation_matrix,columns=target_tokens,index=source_tokens)


    return df,fig2

def get_heatmap_iupac_selfies_all(idx,ids_sat,token_position,dataset,attention,selfies_tokenizer,iupac_tokenizer,atom_vocab_reverse):
    index = ids_sat[idx]
    selfies_sample = dataset[index][1][0]['selfies_tokens']
    iupac_sample = dataset[index][2][0]['iupac_tokens']
    selfies_tokens = selfies_tokenizer.convert_ids_to_tokens(selfies_sample.tolist())   
    mol = dataset[index][3]

  
    iupac_tokens = iupac_tokenizer.convert_ids_to_tokens(iupac_sample.tolist())

    atom_tokens = [atom_vocab_reverse[q] for q in mol.x[:,0].tolist()]
    heatmap_atom = []
    attention_sum = []
    heatmap_name = []


    selfies_tokens_valid = []
    atom_tokens_valid = atom_tokens
    for j in range(1,attention[0].shape[0]):
        if selfies_tokens[j] == '[PAD]':
            break
        selfies_tokens_valid.append(selfies_tokens[j])
    iupac_tokens_valid = []
    for x in range(1,len(iupac_tokens)):
        if iupac_tokens[x] == '[PAD]':
            break
        iupac_tokens_valid.append(iupac_tokens[x])

    smi = sf.decoder("".join(selfies_tokens_valid[0:-1]))
    selfies_tokens_valid.pop(-1)
    iupac_tokens_valid.pop(-1)
    # print(selfies_tokens_valid)
    # print(selfies_tokens_select)
    # print(atom_tokens_valid)
    for k in range(attention.shape[0]): 
        attention_head = attention[k]
        # 对每一行求和，得到每个selfies_token对应的注意力系数
        attention_sum.append(torch.sum(attention_head[1:j-1,len(iupac_tokens):len(iupac_tokens)+len(atom_tokens)],dim=1).cpu())
        attention_name = attention_head[1:j-1,1:x-1]
        attention_atom = attention_head[1:j-1,len(iupac_tokens):len(iupac_tokens)+len(atom_tokens)]
        heatmap_atom.append(attention_atom.cpu())
        heatmap_name.append(attention_name.cpu())
    
    heatmap_atom = np.array(heatmap_atom)
    heatmap_name = np.array(heatmap_name)
    return attention_sum,heatmap_atom,heatmap_name,selfies_tokens_valid,atom_tokens_valid,iupac_tokens_valid,smi



def plot_heatmap_name_all(heatmap_atom,heatmap_name,selfies_tokens_valid,atom_tokens_valid,iupac_tokens_valid,save_path,layer):
    import matplotlib.pyplot as plt
    import seaborn as sns
    import matplotlib.gridspec as gridspec
    # heatmap_name shape (num_heatmaps, rows, columns)


    num_heatmaps = heatmap_name.shape[0]
    # rows = int(np.ceil(num_heatmaps / 2))

    for idx in range(num_heatmaps):
        # 定义一个GridSpec，将画布分为rows行，3列
        fig = plt.figure(figsize=(10,10))
        gs = gridspec.GridSpec(1, 2, width_ratios=[0.95,0.01])


        ax = plt.subplot(gs[0, 0])
        # 转置heatmap_name[idx]，使得行对应于iupac_tokens_valid，列对应于selfies_tokens_valid
        sns.heatmap(heatmap_atom[idx],ax=ax, cmap='PuBu', cbar=False)
        ax.set_title(f'Head {idx + 1}',fontdict={'fontsize': 10, 'fontweight': 'bold','fontfamily':'Operator Mono'})
        ax.set_yticks(np.arange(len(selfies_tokens_valid)) + 0.5, selfies_tokens_valid, rotation=0, fontsize=20)
        ax.set_xticks(np.arange(len(atom_tokens_valid)) + 0.5, atom_tokens_valid, rotation=0, fontsize=20)
        # cloarbar third column
        cax = plt.subplot(gs[:, -1])
        cbar = plt.colorbar(ax.collections[0], cax=cax)
        cbar.ax.set_ylabel('attention score', rotation=90, labelpad=15)  # add colorbar title

        plt.tight_layout()

        fig.savefig(f'{save_path}/layer{layer}_head{idx}_atom.svg',dpi=600,format='svg')
    
    for idx in range(num_heatmaps):
        fig = plt.figure(figsize=(10,10))
        gs = gridspec.GridSpec(1, 2, width_ratios=[0.95,0.01])


        ax = plt.subplot(gs[0, 0])
        # 转置heatmap_name[idx]，使得行对应于iupac_tokens_valid，列对应于selfies_tokens_valid
        sns.heatmap(heatmap_name[idx],ax=ax, cmap='PuBu', cbar=False)
        ax.set_title(f'Head {idx + 1}',fontdict={'fontsize': 10, 'fontweight': 'bold','fontfamily':'Operator Mono'})
        ax.set_yticks(np.arange(len(selfies_tokens_valid)) + 0.5, selfies_tokens_valid, rotation=0, fontsize=20)
        ax.set_xticks(np.arange(len(iupac_tokens_valid)) + 0.5,iupac_tokens_valid, rotation=45, fontsize=15)
        # cloarbar third column
        cax = plt.subplot(gs[:, -1])
        cbar = plt.colorbar(ax.collections[0], cax=cax)
        cbar.ax.set_ylabel('attention score', rotation=90, labelpad=15)  # add colorbar title

        plt.tight_layout()

        fig.savefig(f'{save_path}/layer{layer}_head{idx}_name.svg',dpi=600,format='svg')      
    
    return None


def main(args,config):
    output_dir = args.output_dir
    device = args.device
    iupac_token = args.iupac_token      # a list
    mode = args.mode
    selfies_token = args.selfies_token
    model_name = args.model_name


    print("creating dataset")
    dataset,_ = create_dataset(
    config['dataset']['ids_path'],
    config['dataset']['ipc_path'],
    config['dataset']['sfs_path'],
    config['dataset']['mlg_path'],
    0.0)
    print(len(dataset))
    model = TMMF(config)
    model = model.to(device)

    if model_name == 'esol':
        path = './pre_train/1m_checkpoint_09.pth'
    elif model_name == 'lipo':
        path = './pre_train/lipo_checkpoint.pth'
    elif model_name == 'pre':
        path = './pre_train/1m_checkpoint_09.pth'
    checkpoint = torch.load(path)
    model.load_state_dict(checkpoint['model'])
    Interact_model = BertForInteract(model.selfies_encoder,model.iupac_encoder,model.mol_encoder,model.fusion_encoder)
    Interact_model = Interact_model.to(device)

    iupac_tokenizer = Iupac_tokenizer()
    selfies_tokenizer = Selfies_tokenizer()

    atom_list = ['H','He','Li','Be','B','C','N','O','F','Ne','Na','Mg','Al','Si','P','S','Cl','Ar','K','Ca','Sc','Ti','V','Cr','Mn','Fe','Co',
                 'Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr','Rb','Sr','Y','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd','In','Sn',
                 'Sb','Te','I','Xe']
    atom_dict = {atom:i for i,atom in enumerate(atom_list)}
    atom_dict['<unk>'] = len(atom_dict)
    atom_vocab_reverse = {v:k for k,v in atom_dict.items()}

    if mode == 'iupac':
        ids_sat,token_position = find_satisfy_iupac(dataset,iupac_token,iupac_tokenizer,sample_num=args.sample_num)
        # choose one sample from ids_sat
        for i in range(len(ids_sat)):
            selfies,iupac,mol = dataset[ids_sat[i]][1][0]['selfies_tokens'],dataset[ids_sat[i]][2][0]['iupac_tokens'],dataset[ids_sat[i]][3]
            selfies,iupac = torch.unsqueeze(selfies,dim=0),torch.unsqueeze(iupac,dim=0)
            from torch_geometric.data import Batch
            mol = Batch.from_data_list([mol])
            selfies,iupac,mol = selfies.to(device),iupac.to(device),mol.to(device)
            out = Interact_model(selfies_ids=selfies,iupac_ids=iupac,mol_ids=mol,save_attention=True)
            attention = Interact_model.fusion_encoder.transformer_blocks[args.layer].cross_attention.get_attention_map()[0]
            heatmap_name,selfies_tokens_valid,iupac_tokens_valid,smi = get_heatmap_iupac_selfies(i,ids_sat,token_position,dataset,attention,selfies_tokenizer,iupac_tokenizer)
            df,fig2 = plot_heatmap_name(heatmap_name,selfies_tokens_valid,iupac_tokens_valid)
            save_path = f'{output_dir}/{iupac_token}/{i}'
            if not os.path.exists(save_path):
                os.makedirs(save_path)
            df.to_csv(f'{save_path}/layer{args.layer}.csv')
            # fig1.write_image(f'{save_path}/layer{args.layer}_1.svg')
            # fig1.write_image(f'{save_path}/layer{args.layer}_1.png')
            # fig2.write_image(f'{save_path}/layer{args.layer}_2.svg')
            fig2.write_image(f'{save_path}/layer{args.layer}_2.png')
            smi_graph = Chem.MolFromSmiles(smi)
            from rdkit.Chem.Draw import rdMolDraw2D
            drawer = rdMolDraw2D.MolDraw2DSVG(300, 300)  
            drawer.DrawMolecule(smi_graph,highlightAtoms=[], highlightBonds=[])
            drawer.FinishDrawing()

            svg = drawer.GetDrawingText().replace("svg:", "")

            with open(f'{save_path}/layer{args.layer}_smi.svg', "w") as f:
                f.write(svg)  
            with open(f'{save_path}/layer{args.layer}_smi.txt', "w") as f:
                f.write(smi)

    elif mode == 'atom':
        ids_sat,token_position = find_satify_mol(dataset,selfies_token,selfies_tokenizer,atom_vocab_reverse,sample_num=args.sample_num)
        for i in range(len(ids_sat)):
            selfies,iupac,mol = dataset[ids_sat[i]][1][0]['selfies_tokens'],dataset[ids_sat[i]][2][0]['iupac_tokens'],dataset[ids_sat[i]][3]
            selfies,iupac = torch.unsqueeze(selfies,dim=0),torch.unsqueeze(iupac,dim=0)
            from torch_geometric.data import Batch
            mol = Batch.from_data_list([mol])
            selfies,iupac,mol = selfies.to(device),iupac.to(device),mol.to(device)
            out = Interact_model(selfies_ids=selfies,iupac_ids=iupac,mol_ids=mol,save_attention=True)
            attention = Interact_model.fusion_encoder.transformer_blocks[args.layer].cross_attention.get_attention_map()[0]
            heatmap_atom,selfies_tokens_select,atom_tokens_valid,smi = get_heatmap_mol_selfies(i,ids_sat,token_position,dataset,attention,selfies_tokenizer,iupac_tokenizer,atom_vocab_reverse)
            df,fig2 = plot_heatmap_atom(heatmap_atom,selfies_tokens_select,atom_tokens_valid)
            save_path = f'{output_dir}/{selfies_token}/{i}'
            if not os.path.exists(save_path):
                os.makedirs(save_path)
            df.to_csv(f'{save_path}/layer{args.layer}.csv')
            # fig1.write_image(f'{save_path}/layer{args.layer}_1.svg')
            # fig1.write_image(f'{save_path}/layer{args.layer}_1.png')
            # fig2.write_image(f'{save_path}/layer{args.layer}_2.svg')
            fig2.write_image(f'{save_path}/layer{args.layer}_2.png')
            smi_graph = Chem.MolFromSmiles(smi)
            from rdkit.Chem.Draw import rdMolDraw2D
            drawer = rdMolDraw2D.MolDraw2DSVG(300, 300)  
            drawer.DrawMolecule(smi_graph,highlightAtoms=[], highlightBonds=[])
            drawer.FinishDrawing()

            svg = drawer.GetDrawingText().replace("svg:", "")

            with open(f'{save_path}/layer{args.layer}_smi.svg', "w") as f:
                f.write(svg)  
            with open(f'{save_path}/layer{args.layer}_smi.txt', "w") as f:
                f.write(smi)

    elif mode == 'selfies_iupac':
        iupac_list = iupac_token
        ids_sat,token_position = find_satisfy_iupac_cer(dataset,iupac_list,iupac_tokenizer,sample_num=args.sample_num)
        # choose one sample from ids_sat
        for i in range(len(ids_sat)):
            selfies,iupac,mol = dataset[ids_sat[i]][1][0]['selfies_tokens'],dataset[ids_sat[i]][2][0]['iupac_tokens'],dataset[ids_sat[i]][3]
            selfies,iupac = torch.unsqueeze(selfies,dim=0),torch.unsqueeze(iupac,dim=0)

            from torch_geometric.data import Batch
            mol = Batch.from_data_list([mol])
            selfies,iupac,mol = selfies.to(device),iupac.to(device),mol.to(device)
            out = Interact_model(selfies_ids=selfies,iupac_ids=iupac,mol_ids=mol,save_attention=True)
            attention = Interact_model.fusion_encoder.transformer_blocks[args.layer].cross_attention.get_attention_map()[0]
            try:
                attention_sum,heatmap_atom,heatmap_name,selfies_tokens_valid,atom_tokens_valid,iupac_tokens_valid,smi = get_heatmap_iupac_selfies_all(i,ids_sat,token_position,dataset,attention,selfies_tokenizer,iupac_tokenizer,atom_vocab_reverse)
                # save_path = f'{output_dir}/{iupac_list[i]}_heatmap/{i}'
                save_path = f'{output_dir}/{model_name}/{iupac_list[i]}'
                if not os.path.exists(save_path):
                    os.makedirs(save_path)
                fig = plot_heatmap_name_all(heatmap_atom,heatmap_name,selfies_tokens_valid,atom_tokens_valid,iupac_tokens_valid,save_path,args.layer)
                # 将 attention_atom的8个头分别保存为csv
                import pandas as pd
                # for j in range(len(attention_sum)):
                #     df = pd.DataFrame(attention_sum[j],columns=['atten'],index=selfies_tokens_valid)
                #     df.to_csv(f'{save_path}/layer{args.layer}_head{j}.csv')

                for j in range(heatmap_atom.shape[0]):
                    df = pd.DataFrame(np.array(heatmap_atom[j]),columns=atom_tokens_valid,index=selfies_tokens_valid)
                    df.to_csv(f'{save_path}/layer{args.layer}_head{j}.csv')



                smi_graph = Chem.MolFromSmiles(smi)
                from rdkit.Chem.Draw import rdMolDraw2D
                drawer = rdMolDraw2D.MolDraw2DSVG(300, 300)  
                drawer.DrawMolecule(smi_graph,highlightAtoms=[], highlightBonds=[])
                drawer.FinishDrawing()

                svg = drawer.GetDrawingText().replace("svg:", "")

                with open(f'{save_path}/layer{args.layer}_smi.svg', "w") as f:
                    f.write(svg)  
                # 将 smi 保存为txt
                with open(f'{save_path}/layer{args.layer}_smi.txt', "w") as f:
                    f.write(smi)
            except Exception as e:
                print(e)
                pass

    elif mode == 'iupac_statistics':
        tokens_coefficent_dict = {}
        tokens_count_dict = {}
        ids_sat,token_position = find_satisfy_iupac(dataset,iupac_token,iupac_tokenizer,sample_num=args.sample_num)
        # choose one sample from ids_sat
        topk = 3
        for i in range(len(ids_sat)):
            selfies,iupac,mol = dataset[ids_sat[i]][1][0]['selfies_tokens'],dataset[ids_sat[i]][2][0]['iupac_tokens'],dataset[ids_sat[i]][3]
            selfies,iupac = torch.unsqueeze(selfies,dim=0),torch.unsqueeze(iupac,dim=0)
            from torch_geometric.data import Batch
            mol = Batch.from_data_list([mol])
            selfies,iupac,mol = selfies.to(device),iupac.to(device),mol.to(device)
            out = Interact_model(selfies_ids=selfies,iupac_ids=iupac,mol_ids=mol,save_attention=True)
            attention = Interact_model.fusion_encoder.transformer_blocks[args.layer].cross_attention.get_attention_map()[0]
            heatmap_name,selfies_tokens_valid,iupac_tokens_valid,smi = get_heatmap_iupac_selfies(i,ids_sat,token_position,dataset,attention,selfies_tokenizer,iupac_tokenizer)
            topk_tokens,topk_coefficent = statistics_name_topk(heatmap_name,selfies_tokens_valid,iupac_tokens_valid,topk)
            for j in range(len(topk_tokens)):
                if topk_tokens[j] in tokens_coefficent_dict.keys():
                    tokens_coefficent_dict[topk_tokens[j]] = (tokens_coefficent_dict[topk_tokens[j]]*tokens_count_dict[topk_tokens[j]]+topk_coefficent[j])/(tokens_count_dict[topk_tokens[j]]+1)
                    tokens_count_dict[topk_tokens[j]] += 1
                else:
                    tokens_coefficent_dict[topk_tokens[j]] = topk_coefficent[j]
                    tokens_count_dict[topk_tokens[j]] = 1
            
        # 分为三列 tokens coefficent count
        tokens = []
        coefficent = []
        count = []
        sum_count = sum(tokens_count_dict.values())
        for key in tokens_coefficent_dict.keys():
            tokens.append(key)
            coefficent.append(tokens_coefficent_dict[key])
            count.append(tokens_count_dict[key]/sum_count)
        # 按照 count 大小排序
        tokens = [x for _,x in sorted(zip(count,tokens),reverse=True)]
        coefficent = [x for _,x in sorted(zip(count,coefficent),reverse=True)]
        count = sorted(count,reverse=True)

        import pandas as pd
        df = pd.DataFrame({'tokens':tokens,'coefficent':coefficent,'count':count})
        save_path = f'{output_dir}/{iupac_token}/statistics'
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        df.to_csv(f'{save_path}/top_{topk}.csv')
        
    elif mode == 'atom_statistics':
        tokens_coefficent_dict = {}
        tokens_count_dict = {}
        ids_sat,token_position = find_satify_mol(dataset,selfies_token,selfies_tokenizer,atom_vocab_reverse,sample_num=args.sample_num)
        # choose one sample from ids_sat
        topk = 1
        for i in range(len(ids_sat)):
            selfies,iupac,mol = dataset[ids_sat[i]][1][0]['selfies_tokens'],dataset[ids_sat[i]][2][0]['iupac_tokens'],dataset[ids_sat[i]][3]
            selfies,iupac = torch.unsqueeze(selfies,dim=0),torch.unsqueeze(iupac,dim=0)
            from torch_geometric.data import Batch
            mol = Batch.from_data_list([mol])
            selfies,iupac,mol = selfies.to(device),iupac.to(device),mol.to(device)
            out = Interact_model(selfies_ids=selfies,iupac_ids=iupac,mol_ids=mol,save_attention=True)
            attention = Interact_model.fusion_encoder.transformer_blocks[args.layer].cross_attention.get_attention_map()[0]
            heatmap_atom,selfies_tokens_select,atom_tokens_valid,smi = get_heatmap_mol_selfies(i,ids_sat,token_position,dataset,attention,selfies_tokenizer,iupac_tokenizer,atom_vocab_reverse)
            topk_tokens,topk_coefficent = statistics_atom_topk(heatmap_atom,selfies_tokens_select,atom_tokens_valid,topk)
            for j in range(len(topk_tokens)):
                if topk_tokens[j] in tokens_coefficent_dict.keys():
                    tokens_coefficent_dict[topk_tokens[j]] = (tokens_coefficent_dict[topk_tokens[j]]*tokens_count_dict[topk_tokens[j]]+topk_coefficent[j])/(tokens_count_dict[topk_tokens[j]]+1)
                    tokens_count_dict[topk_tokens[j]] += 1
                else:
                    tokens_coefficent_dict[topk_tokens[j]] = topk_coefficent[j]
                    tokens_count_dict[topk_tokens[j]] = 1
        
        tokens = []
        coefficent = []
        count = []
        sum_count = sum(tokens_count_dict.values())
        for key in tokens_coefficent_dict.keys():
            tokens.append(key)
            coefficent.append(tokens_coefficent_dict[key])
            count.append(tokens_count_dict[key]/sum_count)
        # 按照 count 大小排序
        tokens = [x for _,x in sorted(zip(count,tokens),reverse=True)]
        coefficent = [x for _,x in sorted(zip(count,coefficent),reverse=True)]
        count = sorted(count,reverse=True)
        import pandas as pd
        df = pd.DataFrame({'tokens':tokens,'coefficent':coefficent,'count':count})
        save_path = f'{output_dir}/{selfies_token}/statistics'
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        df.to_csv(f'{save_path}/top_{topk}.csv')
        
if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', default='./config/retrieval.yaml')
    parser.add_argument('--layer', default=0)
    parser.add_argument('--device', default='cuda:0')
    parser.add_argument('--output_dir', default='downstream/interact')
    parser.add_argument('--iupac_token',type=str, nargs='+', default=['formyl'])
    parser.add_argument('--model_name',type=str, default='pre')
    parser.add_argument('--selfies_token', default=['[#N]','[Cl]','[=O]'])
    parser.add_argument('--mode', default='iupac_statistics')
    parser.add_argument('--sample_num', default=2000)
    args = parser.parse_args()

    config = yaml.load(open(args.config, 'r'), Loader=yaml.Loader)

     
    main(args, config)
    
 



"""
BERT Encoder
"""

import torch
from torch import nn
import math
import torch.nn.functional as F


class BertEmbedding(nn.Module):

    def __init__(self,config):
        super(BertEmbedding,self).__init__()
        self.token_embedding = nn.Embedding(config.vocab_size,config.hidden_size,padding_idx=0)
        self.position_embedding = nn.Embedding(config.max_position_embeddings,config.hidden_size)
        self.token_type_embedding = nn.Embedding(config.type_vocab_size,config.hidden_size)

        self.LayerNorm = nn.LayerNorm(config.hidden_size,eps=1e-12)
        self.dropout = nn.Dropout(config.dropout_rate)
        self.config = config

    
    def forward(self,input_ids,token_type_ids=None):
        seq_length = input_ids.size(1)
        position_ids = torch.arange(seq_length,dtype=torch.long,device=input_ids.device)
        position_ids = position_ids.unsqueeze(0).expand_as(input_ids)

        if token_type_ids is None:
            token_type_ids = torch.zeros_like(input_ids)

        token_embeddings = self.token_embedding(input_ids)
        position_embeddings = self.position_embedding(position_ids)
        token_type_embeddings = self.token_type_embedding(token_type_ids)

        embeddings = token_embeddings + position_embeddings + token_type_embeddings
        embeddings = self.LayerNorm(embeddings)
        embeddings = self.dropout(embeddings)

        return embeddings


class GELU(nn.Module):
    """
    Paper Section 3.4, last paragraph notice that BERT used the GELU instead of RELU
    """

    def forward(self, x):
        return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
    
class SublayerConnection(nn.Module):

    def __init__(self, size, dropout):
        super(SublayerConnection, self).__init__()
        self.norm = nn.LayerNorm(size)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x, sublayer):

        return x + self.dropout(sublayer(self.norm(x)))    

class PositionwiseFeedForward(nn.Module):
    "Implements FFN equation."

    def __init__(self, d_model, d_ff, dropout=0.1):
        super(PositionwiseFeedForward, self).__init__()
        self.w_1 = nn.Linear(d_model, d_ff)
        self.w_2 = nn.Linear(d_ff, d_model)
        self.dropout = nn.Dropout(dropout)
        self.activation = GELU()

    def forward(self, x):
        return self.w_2(self.dropout(self.activation(self.w_1(x))))
    

"""attention """
class Attention(nn.Module):
    """
    Compute 'Scaled Dot Product Attention
    """

    def forward(self, query, key, value, mask=None, dropout=None):
        scores = torch.matmul(query, key.transpose(-2, -1)) \
                 / math.sqrt(query.size(-1))

        if mask is not None:
            scores = scores.masked_fill(mask == 0, -1e9)

        p_attn = F.softmax(scores, dim=-1)

        if dropout is not None:
            p_attn = dropout(p_attn)

        return torch.matmul(p_attn, value), p_attn

"""multiHeadAttention"""
class MultiHeadedAttention(nn.Module):
    """
    Take in model size and number of heads.
    """

    def __init__(self, h, d_model, dropout=0.1):
        super().__init__()
        assert d_model % h == 0

        # We assume d_v always equals d_k
        self.d_k = d_model // h
        self.h = h

        self.linear_layers = nn.ModuleList([nn.Linear(d_model, d_model) for _ in range(3)])
        self.output_linear = nn.Linear(d_model, d_model)
        self.attention = Attention()

        self.dropout = nn.Dropout(p=dropout)

    def save_attention_map(self, attention_map):
        self.attention_map = attention_map
        
    def get_attention_map(self):
        return self.attention_map

    def forward(self, query, key, value, mask=None,save_attention=False):
        batch_size = query.size(0)

        # 1) Do all the linear projections in batch from d_model => h x d_k
        query, key, value = [l(x).view(batch_size, -1, self.h, self.d_k).transpose(1, 2)
                             for l, x in zip(self.linear_layers, (query, key, value))]

        # 2) Apply attention on all the projected vectors in batch.
        x, attn = self.attention(query, key, value, mask=mask, dropout=self.dropout)

        # 3) "Concat" using a view and apply a final linear.
        x = x.transpose(1, 2).contiguous().view(batch_size, -1, self.h * self.d_k)

        if save_attention:
            self.save_attention_map(attn)

        return self.output_linear(x)
    

class TransformerBlock(nn.Module):
    """
    Bidirectional Encoder = Transformer (self-attention)
    Transformer = MultiHead_Attention + Feed_Forward with sublayer connection
    """

    def __init__(self, hidden, attn_heads, feed_forward_hidden, dropout, save_attention=False):
        """
        :param hidden: hidden size of transformer
        :param attn_heads: head sizes of multi-head attention
        :param feed_forward_hidden: feed_forward_hidden, usually 4*hidden_size
        :param dropout: dropout rate
        """

        super().__init__()
        self.attention = MultiHeadedAttention(h=attn_heads, d_model=hidden)
        self.feed_forward = PositionwiseFeedForward(d_model=hidden, d_ff=feed_forward_hidden, dropout=dropout)
        self.input_sublayer = SublayerConnection(size=hidden, dropout=dropout)
        self.output_sublayer = SublayerConnection(size=hidden, dropout=dropout)
        self.dropout = nn.Dropout(p=dropout)
        self.save_attention = save_attention

    def forward(self,x,y,mask,save_attention=False):
        # x is query, y is key and value
        x = self.input_sublayer(x, lambda _x: self.attention.forward(_x, y, y, mask=mask,save_attention=save_attention))
        

        x = self.output_sublayer(x, self.feed_forward)
        return self.dropout(x)


class FusionBlock(nn.Module):
    """
    Bidirectional Encoder = Transformer (self-attention)
    Transformer = MultiHead_Attention + Feed_Forward with sublayer connection
    """

    def __init__(self, hidden, attn_heads, feed_forward_hidden, dropout, save_attention=False):
        """
        :param hidden: hidden size of transformer
        :param attn_heads: head sizes of multi-head attention
        :param feed_forward_hidden: feed_forward_hidden, usually 4*hidden_size
        :param dropout: dropout rate
        """

        super().__init__()
        self.attention = MultiHeadedAttention(h=attn_heads, d_model=hidden)
        self.cross_attention = MultiHeadedAttention(h=attn_heads, d_model=hidden)
        self.feed_forward = PositionwiseFeedForward(d_model=hidden, d_ff=feed_forward_hidden, dropout=dropout)
        self.input_sublayer = SublayerConnection(size=hidden, dropout=dropout)
        self.output_sublayer = SublayerConnection(size=hidden, dropout=dropout)
        self.dropout = nn.Dropout(p=dropout)
        self.save_attention = save_attention

    def forward(self,x,y,mask_self,mask_cross,save_attention=False):
        # x is query, y is key and value
        x = self.input_sublayer(x, lambda _x: self.attention.forward(_x, _x, _x, mask=mask_self,save_attention=False))
        x = self.input_sublayer(x, lambda _x: self.cross_attention.forward(_x, y, y, mask=mask_cross,save_attention=save_attention))        

        x = self.output_sublayer(x, self.feed_forward)
        return self.dropout(x) 

class bert_encoder(nn.Module):

    def __init__(self,config):
        
        super().__init__()
        self.config = config
        self.hidden_size = config.hidden_size
        self.n_layers = config.n_layers
        self.attn_heads = config.attn_heads
        self.vocab_size = config.vocab_size
        self.dropout = config.dropout_rate
        self.mode = config.mode

        self.embedding = BertEmbedding(config)

        if self.mode == 'fusion':
            self.transformer_blocks =  nn.ModuleList(
                [FusionBlock(self.hidden_size, self.attn_heads, self.hidden_size * 4, self.dropout) for _ in range(self.n_layers)])

        else:
            self.transformer_blocks = nn.ModuleList(
                [TransformerBlock(self.hidden_size, self.attn_heads, self.hidden_size * 4, self.dropout) for _ in range(self.n_layers)])
            

    
    def forward(self, input_ids=None,selfies_embeds=None,iupac_mol_embeds=None,selfies_mask=None,iupac_mol_mask=None,mode=None,save_attention=False):

        mode = self.mode if mode is None else mode
        if mode == 'iupac' or mode=='selfies':
            x = input_ids

            mask = (x > 0)
            mask_attn = mask.unsqueeze(1).repeat(1, mask.size(1), 1).unsqueeze(1).to(input_ids.device)
            x = self.embedding(x)
            for transformer in self.transformer_blocks:
                x = transformer.forward(x,x,mask_attn)

            return x,mask
        
        if mode == 'fusion':
            # y consit of both iuapc and molecular graph features
            self.transformer_blocks[0].save_attention = True
            mask_selfies = selfies_mask.clone()
            mask_selfies = mask_selfies.unsqueeze(2).to(torch.float32)
            mask_iupac_mol = iupac_mol_mask.unsqueeze(1).to(torch.float32)
            mask_cross = torch.matmul(mask_selfies,mask_iupac_mol).unsqueeze(1).to(selfies_embeds.device)
            mask_self = selfies_mask.unsqueeze(1).repeat(1, selfies_mask.size(1), 1).unsqueeze(1).to(selfies_embeds.device)
            i = 0
            for transformer in self.transformer_blocks:
                selfies_embeds = transformer.forward(selfies_embeds,iupac_mol_embeds,mask_self,mask_cross,save_attention=True)
                # if save_attention and transformer.save_attention:
                #     torch.save(transformer.cross_attention.get_attention_map(),'./attention/cross_attention_'+ str(i) +'.pt')
                #     i += 1
            return selfies_embeds
        
        return None
    
class BertForMaskSM(nn.Module):

    def __init__(self,config,selfies_encoder,iupac_encoder,mol_encoder,fusion_encoder):
        super().__init__()
        self.config = config
        self.selfies_encoder = selfies_encoder
        self.iupac_encoder = iupac_encoder
        self.mol_encoder = mol_encoder
        self.fusion_encoder = fusion_encoder
        self.fc = nn.Linear(config.hidden_size,config.vocab_size)
        self.softmax = nn.LogSoftmax(dim=-1)

    def forward(self,selfies_ids=None,iupac_ids=None,mol_ids=None,labels=None,mode=None):

        selfies_embeds,selfies_mask = self.selfies_encoder(input_ids=selfies_ids,mode='selfies')
        iupac_embeds,iupac_mask = self.iupac_encoder(input_ids=iupac_ids,mode='iupac')
        mol_pool,mol_embeds,mol_mask = self.mol_encoder(mol_ids)

        iupac_mol_embeds = torch.cat([iupac_embeds,mol_embeds],dim=1)
        iupac_mol_mask = torch.cat([iupac_mask,mol_mask],dim=1)
        output = self.fusion_encoder(selfies_embeds=selfies_embeds,
                                     iupac_mol_embeds=iupac_mol_embeds,
                                     selfies_mask=selfies_mask,
                                     iupac_mol_mask=iupac_mol_mask,
                                     mode='fusion')
        if mode == 'output':
            return output[:,:,:]
        output = self.softmax(self.fc(output))

        if labels is not None:
            loss_fct = nn.NLLLoss(ignore_index=0)
            loss_msm = loss_fct(output.transpose(1, 2),labels)

            if mode == 'train':
                return output,loss_msm
            
            if mode == 'test':
                pred = output.transpose(1, 2).argmax(dim=1)
                num = torch.ne(labels,0).sum().item()
                for k in range(len(labels)):
                    for j in range(len(labels[k])):
                        if labels[k][j]==0: labels[k][j] = -100
                correct = torch.eq(pred,labels).sum().float().item()
                return correct,num
            
        else:
            return output,None
        

class BertForFinetune(nn.Module):

    def __init__(self,config,selfies_encoder,iupac_encoder,mol_encoder,fusion_encoder):
        super().__init__()
        self.config = config
        self.selfies_encoder = selfies_encoder
        self.iupac_encoder = iupac_encoder
        self.mol_encoder = mol_encoder
        self.fusion_encoder = fusion_encoder
        self.task_type = config.dataset.task_type
        self.metrics = config.dataset.metrics
        self.dropuout = nn.Dropout(config.dropout)
        # self.softmax = nn.LogSoftmax(dim=-1)
        if self.task_type == 'classification':
            # for p in self.parameters():
            #     p.requires_grad=False
            self.num_task = config.dataset.num_task
            self.linear1 = nn.Linear(config.hidden_size,config.hidden_size)
            self.activation = nn.Tanh()
            self.linear2 = nn.Linear(config.hidden_size,config.dataset.num_task)
        
        if self.task_type == 'regression':
            self.linear1 = nn.Linear(config.hidden_size,config.hidden_size)
            self.activation = nn.Tanh()
            self.linear2 = nn.Linear(config.hidden_size,1)


    def forward(self,selfies_ids=None,iupac_ids=None,mol_ids=None,labels=None):
        selfies_embeds,selfies_mask = self.selfies_encoder(input_ids=selfies_ids,mode='selfies')
        iupac_embeds,iupac_mask = self.iupac_encoder(input_ids=iupac_ids,mode='iupac')
        mol_pool,mol_embeds,mol_mask = self.mol_encoder(mol_ids)

        iupac_mol_embeds = torch.cat([iupac_embeds,mol_embeds],dim=1)
        iupac_mol_mask = torch.cat([iupac_mask,mol_mask],dim=1)
        output = self.fusion_encoder(selfies_embeds=selfies_embeds,
                                     iupac_mol_embeds=iupac_mol_embeds,
                                     selfies_mask=selfies_mask,
                                     iupac_mol_mask=iupac_mol_mask,
                                     mode='fusion')
        output = self.linear1(output[:,0,:])
        output = self.activation(output)
        output = self.linear2(self.dropuout(output))

        return output



    def _train(self,selfies_ids=None,iupac_ids=None,mol_ids=None,labels=None):
        output = self.forward(selfies_ids,iupac_ids,mol_ids)

        if self.task_type == 'classification':
            m = nn.Sigmoid()
            pred = m(output)
            loss_fct = nn.BCELoss(reduction = "none")
            is_valid = labels**2 > 0 
            loss_mat = loss_fct(pred,(labels+1)/2)
            loss_mat = torch.where(is_valid, loss_mat, torch.zeros(loss_mat.shape).to(loss_mat.device).to(loss_mat.dtype))
            loss = torch.sum(loss_mat)/torch.sum(is_valid)
            return pred,loss
        elif self.task_type == 'regression':
            loss_fct = nn.MSELoss()
            loss = loss_fct(output.squeeze(-1),labels)
            return output,loss
        
        return None,None
    

    def _test(self,selfies_ids=None,iupac_ids=None,mol_ids=None,labels=None):
        output = self.forward(selfies_ids,iupac_ids,mol_ids)

        if self.task_type == 'classification':
            m = nn.Sigmoid()
            pred = m(output)
            return pred,labels
        
        if self.task_type == 'regression':
            pred = output.squeeze(-1)
            return pred,labels





        






from .dataset.data4csv import Data4CSV
from .dataset.data4csv import Data4CSV_collate
from .dataset.data4json import JsonDataset,JsonDataset_collate
from torch.utils.data import DataLoader
import torch
from .model.TMMF import TMMF
import ruamel.yaml as yaml
import os
from tqdm import tqdm
from dataclasses import field
from rdkit.Chem import AllChem



def get_data_loader_csv(data_path, batch_size, shuffle=False):
    Data4CSV_dataset = Data4CSV(data_path, selfies_max_len=130, iupac_max_len=130)
    dataloader = DataLoader(
        Data4CSV_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        collate_fn=Data4CSV_collate,
        drop_last=False  
    )

    return dataloader

def get_data_loader_json(data_path, batch_size, shuffle=False):
    Json_dataset = JsonDataset(data_path, selfies_max_len=130, iupac_max_len=130)
    dataloader = DataLoader(
        Json_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        collate_fn=JsonDataset_collate,
        drop_last=False  
    )

    return dataloader

def get_model(model_path: str = field(default = None), 
              device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
              config_path:str = './config/pre_train.yaml'):


    config = yaml.load(open(config_path, 'r'), Loader=yaml.Loader)

    print("Creating model...")
    model = TMMF(config)
    model = model.to(device)

    # path = './pre_train/1m_checkpoint_09.pth'
    checkpoint = torch.load(model_path, map_location=device)
    model.load_state_dict(checkpoint['model'])
    model.eval()

    return model

class BatchFeature:
    def __init__(self, model_path: str = None, device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu"), config_path:str = './config/pre_train.yaml'):
        self.model = get_model(model_path, device, config_path)
        self.device = device
    
    def get_feature_single(self, smiles:str, iupac_name:str, selfies:str=None, mlg_ids=None):
        from .dataset.tokenizer import Iupac_tokenizer,Selfies_tokenizer
        from .dataset.atom_rep import mol_to_graph_data_obj_simple
        ipc_tokenizer = Iupac_tokenizer()
        ipc_ids = ipc_tokenizer.convert_tokens_to_ids(ipc_tokenizer.tokenize(iupac_name))
        ipc_ids = [ipc_tokenizer.convert_tokens_to_ids(['[CLS]'])[0]] + ipc_ids + [ipc_tokenizer.convert_tokens_to_ids(['[EOS]'])[0]]
        if selfies is None:
            import selfies as sf
            selfies = sf.encoder(smiles)
        sfs_tokenizer = Selfies_tokenizer()
        sfs_ids = sfs_tokenizer.convert_tokens_to_ids(sfs_tokenizer.tokenize(selfies))
        sfs_ids = [sfs_tokenizer.convert_tokens_to_ids(['[CLS]'])[0]] + sfs_ids + [sfs_tokenizer.convert_tokens_to_ids(['[EOS]'])[0]]

        rdkit_mol = AllChem.MolFromSmiles(smiles)
        if rdkit_mol != None:  # ignore invalid mol objects
            mlg_ids = mol_to_graph_data_obj_simple(rdkit_mol)
        
        ipc_ids = [ipc_ids]
        sfs_ids = [sfs_ids]

        from torch_geometric.data import Batch
        mlg_ids = Batch.from_data_list([mlg_ids]).to(self.device)
        ipc_ids = torch.tensor(ipc_ids, dtype=torch.long).to(self.device)
        sfs_ids = torch.tensor(sfs_ids, dtype=torch.long).to(self.device)

        out = self.model.get_fusion_feature(mlg_ids,ipc_ids,sfs_ids)

        return out

    def get_feature(self, data_path:str=None, batch_size:int=32,output_dir:str=None):
        if data_path.endswith('.csv'):
            dataloader = get_data_loader_csv(data_path, batch_size, shuffle=False)
        elif data_path.endswith('.json'):
            dataloader = get_data_loader_json(data_path, batch_size, shuffle=False)
        else:
            raise ValueError('Data path should be a csv or json file')
        os.makedirs(output_dir, exist_ok=True)


        for index,batch in enumerate(tqdm(dataloader, desc="Processing batches", unit="batch")):
            if data_path.endswith('.csv'):
                cids = batch['cids']
            elif data_path.endswith('.json'):
                cids = batch['ids']
            selfies_tokens = batch['selfies_tokens'].to(self.device)  
            iupac_tokens = batch['iupac_tokens'].to(self.device) 
            mol_graph = batch['mol_graph'].to(self.device) 
            with torch.no_grad():
                out = self.model.get_fusion_feature(mol_graph, iupac_tokens, selfies_tokens)

            # 将输出保存为文件
            for i, cid in enumerate(cids):
                # 提取当前样本的输出
                output_tensor = out[i].cpu()  # 将张量移动到 CPU
                # 保存为文件
                output_path = os.path.join(output_dir, f'{cid}.pt')
                torch.save(output_tensor, output_path)
            



if __name__ == '__main__':
    model_path = './pre_train/1m_checkpoint_09.pth'
    data_path = '/home/hukun/code/trimodal_code/craft_mol/data/Mol-Instructions/data/Molecule-oriented_Instructions/molecular_description_generation/test_data.json'
    output_dir = '/home/hukun/code/trimodal_code/craft_mol/data/Mol-Instructions/data/Molecule-oriented_Instructions/molecular_description_generation/test_output_features'
    batch_size = 32

    batch_feature = BatchFeature(model_path)
    batch_feature.get_feature(data_path, batch_size, output_dir)




import pandas as pd
import glob

# 定义文件名模式
file_pattern = "./data/pubchem324k/pubchem_data_*.csv"

# 初始化一个空列表，用于存储每个文件的 DataFrame
dataframes = []

# 遍历所有匹配的文件
for file_name in glob.glob(file_pattern):
    # 读取 CSV 文件
    print(file_name)
    # df = pd.read_csv(file_name)
    # 将 DataFrame 添加到列表中
    # dataframes.append(df)

# 合并所有 DataFrame
# combined_df = pd.concat(dataframes, ignore_index=True)

# # 保存合并后的 DataFrame 到一个新的 CSV 文件
# combined_df.to_csv("combined_pubchem_data.csv", index=False)

# print("合并完成！合并后的文件已保存为 combined_pubchem_data.csv")


import torch
import pandas as pd
from rdkit.Chem import AllChem
from torch_geometric.data import Data
from torch.utils.data import Dataset
import selfies as sf
from .tokenizer import Iupac_tokenizer, Selfies_tokenizer
from .atom_rep import mol_to_graph_data_obj_simple


class Data4CSV(Dataset):
    def __init__(self, csv_path, selfies_max_len, iupac_max_len):
        """
        Initialize the dataset.
        
        Parameters:
            csv_path (str): CSV path
            selfies_max_len (int): SELFIES max length
            iupac_max_len (int): IUPAC max length
        """
        self.selfies_max_len = selfies_max_len
        self.iupac_max_len = iupac_max_len
        

        self.data = pd.read_csv(csv_path)
        

        self.iupac_tokenizer = Iupac_tokenizer()
        self.selfies_tokenizer = Selfies_tokenizer()
        
 
        self.iupac_cls_token = self.iupac_tokenizer.convert_tokens_to_ids(['[CLS]'])[0]
        self.iupac_eos_token = self.iupac_tokenizer.convert_tokens_to_ids(['[EOS]'])[0]
        
        self.selfies_cls_token = self.selfies_tokenizer.convert_tokens_to_ids(['[CLS]'])[0]
        self.selfies_eos_token = self.selfies_tokenizer.convert_tokens_to_ids(['[EOS]'])[0]

    def __len__(self):
        """return the length of the dataset"""
        return len(self.data)

    def __getitem__(self, idx):
        """
        Get the data of the given index.
        
        return:
            dict: The data of the given index.

        """

        row = self.data.iloc[idx]
        cid = int(row['CID'])
        smiles = row['SMILES']
        selfies = None
        try:
            selfies = sf.encoder(smiles)
        except:
            pass
        iupac_name = row['iupac name']
        
        # Initialize the output
        output = {
            'cid': cid,
            'selfies_tokens': None,
            'iupac_tokens': None,
            'mol_graph': None
        }
        
        # handle IUPAC
        if pd.notna(iupac_name):
            output['iupac_tokens'] = self._get_iupac_tokens(iupac_name)
        
        # handle SELFIES
        if pd.notna(selfies):
            output['selfies_tokens'] = self._get_selfies_tokens(selfies)
        
        # handle SMILES
        if pd.notna(smiles):
            output['mol_graph'] = self._get_mol_graph(smiles)
        
        return output

    def _get_iupac_tokens(self, text):
        """convert IUPAC name to tokens."""
        tokens = None
        try:
            tokens = self.iupac_tokenizer.tokenize(text)
            tokens = self.iupac_tokenizer.convert_tokens_to_ids(tokens)
            tokens = self._get_padded_tokens(tokens, self.iupac_max_len, self.iupac_cls_token, self.iupac_eos_token)
        except:
            pass
        return tokens

    def _get_selfies_tokens(self, text):
        """convert SELFIES to tokens."""
        tokens = None
        try:
            tokens = self.selfies_tokenizer.tokenize(text)
            tokens = self.selfies_tokenizer.convert_tokens_to_ids(tokens)
            tokens = self._get_padded_tokens(tokens, self.selfies_max_len, self.selfies_cls_token, self.selfies_eos_token)
        except:
            pass
        return tokens

    def _get_mol_graph(self, smiles):
        """convert SMILES to graph data object."""
        mol = AllChem.MolFromSmiles(smiles)
        if mol is not None:
            return mol_to_graph_data_obj_simple(mol)
        return None

    def _get_padded_tokens(self, tokens, max_len, cls_token, eos_token):
        """padding the tokens to the given length."""
        tokens = tokens[:max_len]
        tokens = [cls_token] + tokens + [eos_token]
        tokens = torch.tensor(tokens, dtype=torch.long)
        
        """padding the tokens to the given length."""
        padded_tokens = torch.zeros(max_len + 2, dtype=torch.long)
        padded_tokens[:len(tokens)] = tokens
        return padded_tokens


from toolz.sandbox import unzip
from torch_geometric.data import Batch

def Data4CSV_collate(inputs):
    """
    Collate function for the Data4CSV dataset.
    """
    cid_list = []
    selfies_tokens_list = []
    iupac_tokens_list = []
    mol_graph_list = []
    
    for sample in inputs:

        if sample['selfies_tokens'] is None or sample['iupac_tokens'] is None or sample['mol_graph'] is None:
            continue  # 跳过当前样本
        
        cid_list.append(sample['cid'])
        selfies_tokens_list.append(sample['selfies_tokens'])
        iupac_tokens_list.append(sample['iupac_tokens'])
        mol_graph_list.append(sample['mol_graph'])
    

    if len(selfies_tokens_list) > 0:
        selfies_tokens_collate = torch.stack(selfies_tokens_list)  
    else:
        selfies_tokens_collate = None
    
    if len(iupac_tokens_list) > 0:
        iupac_tokens_collate = torch.stack(iupac_tokens_list)  
    else:
        iupac_tokens_collate = None
    
    if len(mol_graph_list) > 0:
        mol_graph_collate = Batch.from_data_list(mol_graph_list)  
    else:
        mol_graph_collate = None
    
    

    batch = {
        'cids': cid_list,  # CID list
        'selfies_tokens': selfies_tokens_collate,  # SELFIES tokens
        'iupac_tokens': iupac_tokens_collate,  # IUPAC tokens
        'mol_graph': mol_graph_collate  # molecule graph 
    }
    
    return batch


import torch
import torch.nn.functional as F
from torch import nn

from .bert import BertForMaskSM,bert_encoder
from .gnn import GNN_graphpred

import json


class Config(dict):
    def __init__(self, *args, **kwargs):
        super(Config, self).__init__(*args, **kwargs)

    def __getattr__(self, key):
        value = self[key]
        if isinstance(value, dict):
            value = Config(value)
        return value
                                                 
class TMMF(nn.Module):
    def __init__(self,config=None):
        super().__init__()
        

        self.msm_pro = config['msm_probability']
        
        mol_config = Config(json.load(open(config['gnn_config'])))
        iuapc_config = Config(json.load(open(config['iupac_config'])))
        selfies_config = Config(json.load(open(config['selfies_config'])))
        fusion_config = Config(json.load(open(config['fusion_config'])))
        train_config = Config(json.load(open(config['train_config'])))
        

        self.mol_encoder = GNN_graphpred(mol_config)
        self.iupac_encoder = bert_encoder(iuapc_config) 
        self.selfies_encoder = bert_encoder(selfies_config)
        self.fusion_encoder = bert_encoder(fusion_config)

        self.BertMaskSelfies = BertForMaskSM(config=train_config,
                                             selfies_encoder=self.selfies_encoder,
                                             iupac_encoder=self.iupac_encoder,
                                             mol_encoder=self.mol_encoder,
                                             fusion_encoder=self.fusion_encoder)


        mol_width = config['mol_width']
        iuapc_width = self.iupac_encoder.config.hidden_size
        selfies_width = self.selfies_encoder.config.hidden_size

        embed_dim = config['embed_dim']   # projection dimension


        self.mol_proj = nn.Linear(mol_width, embed_dim)
        self.iupac_proj = nn.Linear(iuapc_width, embed_dim)
        self.selfies_proj = nn.Linear(selfies_width, embed_dim)

        self.temp = nn.Parameter(torch.ones([]) * config['temp'])
        self.quene_size = config['quene_size']
        self.momentum = config['momentum']
        self.sim_head = nn.Linear(selfies_width, 2)


        # create momentum models

        self.mol_encoder_m = GNN_graphpred(mol_config)
        self.iupac_encoder_m = bert_encoder(iuapc_config)
        self.selfies_encoder_m = bert_encoder(selfies_config)
        self.fusion_encoder_m = bert_encoder(fusion_config)

        self.mol_proj_m = nn.Linear(mol_width, embed_dim)
        self.iupac_proj_m = nn.Linear(iuapc_width, embed_dim)
        self.selfies_proj_m = nn.Linear(selfies_width, embed_dim)

        self.model_pairs = [[self.mol_encoder, self.mol_encoder_m],
                            [self.iupac_encoder, self.iupac_encoder_m],
                            [self.selfies_encoder, self.selfies_encoder_m],
                            [self.mol_proj, self.mol_proj_m],
                            [self.iupac_proj, self.iupac_proj_m],
                            [self.selfies_proj, self.selfies_proj_m],
                            [self.fusion_encoder, self.fusion_encoder_m]]
        self.copy_params()

        # crerate the queue

        self.register_buffer("mol_queue", torch.randn(embed_dim, self.quene_size))
        self.register_buffer("iupac_queue", torch.randn(embed_dim, self.quene_size))
        self.register_buffer("selfies_queue", torch.randn(embed_dim, self.quene_size))
        self.register_buffer("queue_ptr", torch.zeros(1, dtype=torch.long))

        self.mol_queue = F.normalize(self.mol_queue, dim=0)
        self.iupac_queue = F.normalize(self.iupac_queue, dim=0)
        self.selfies_queue = F.normalize(self.selfies_queue, dim=0)


    def forward(self, mol, iupac, selfies,alpha=0):
        with torch.no_grad():
            self.temp.clamp_(0.001,0.5)


        iupac_embeds,iupac_mask = self.iupac_encoder(input_ids=iupac)
        iupac_feat = F.normalize(self.iupac_proj(iupac_embeds[:,0,:]), dim=-1)

        selfies_embeds,selfies_mask = self.selfies_encoder(input_ids = selfies)
        selfies_feat = F.normalize(self.selfies_proj(selfies_embeds[:,0,:]), dim=-1)

        mol_pool,mol_embeds,mol_mask = self.mol_encoder(mol)

        mol_feat = F.normalize(self.mol_proj(mol_pool), dim=-1)
    

        # momentum update
        with torch.no_grad():
            self._momentum_update()
            selfies_embeds_m,_ = self.selfies_encoder_m(input_ids = selfies)
            selfies_feat_m = F.normalize(self.selfies_proj_m(selfies_embeds_m[:,0,:]), dim=-1)
            selfies_feat_all = torch.cat([selfies_feat_m.t(),self.selfies_queue.clone().detach()],dim=1)

            iupac_embeds_m,_ = self.iupac_encoder_m(input_ids=iupac)
            iupac_feat_m = F.normalize(self.iupac_proj_m(iupac_embeds_m[:,0,:]), dim=-1)
            iupac_feat_all = torch.cat([iupac_feat_m.t(),self.iupac_queue.clone().detach()],dim=1)

            mol_pool_m,mol_embeds_m,mask_mol_m = self.mol_encoder_m(mol)
            mol_feat_m = F.normalize(self.mol_proj_m(mol_pool_m), dim=-1)
            mol_feat_all = torch.cat([mol_feat_m.t(),self.mol_queue.clone().detach()],dim=1)

            sim_s2i_m = selfies_feat_m @ iupac_feat_all / self.temp
            sim_s2m_m = selfies_feat_m @ mol_feat_all / self.temp
            sim_s2im_m =  (selfies_feat_m @ iupac_feat_all + selfies_feat_m @ mol_feat_all) / self.temp

            smi_i2s_m = iupac_feat_m @ selfies_feat_all / self.temp
            smi_m2s_m = mol_feat_m @ selfies_feat_all / self.temp
            smi_im2s_m = (iupac_feat_m @ selfies_feat_all + mol_feat_m @ selfies_feat_all) / self.temp

            sim_targets = torch.zeros(sim_s2i_m.size()).to(selfies.device)
            sim_targets.fill_diagonal_(1)

            sim_s2i_targets = alpha * F.softmax(sim_s2i_m,dim=1) + (1-alpha) * sim_targets
            sim_s2m_targets = alpha * F.softmax(sim_s2m_m,dim=1) + (1-alpha) * sim_targets
            sim_s2im_targets = alpha * F.softmax(sim_s2im_m,dim=1) + (1-alpha) * sim_targets
            sim_i2s_targets = alpha * F.softmax(smi_i2s_m,dim=1) + (1-alpha) * sim_targets
            sim_m2s_targets = alpha * F.softmax(smi_m2s_m,dim=1) + (1-alpha) * sim_targets
            sim_im2s_targets = alpha * F.softmax(smi_im2s_m,dim=1) + (1-alpha) * sim_targets

        sim_s2i = selfies_feat @ iupac_feat_all / self.temp
        sim_i2s = iupac_feat @ selfies_feat_all / self.temp

        sim_s2m = selfies_feat @ mol_feat_all / self.temp
        sim_m2s = mol_feat @ selfies_feat_all / self.temp

        sim_s2im = (selfies_feat @ iupac_feat_all + selfies_feat @ mol_feat_all) / self.temp
        sim_im2s = (iupac_feat @ selfies_feat_all + mol_feat @ selfies_feat_all) / self.temp

        loss_s2i = -torch.sum(F.log_softmax(sim_s2i,dim=1)*sim_s2i_targets,dim=1).mean()
        loss_i2s = -torch.sum(F.log_softmax(sim_i2s,dim=1)*sim_i2s_targets,dim=1).mean()
        loss_s2m = -torch.sum(F.log_softmax(sim_s2m,dim=1)*sim_s2m_targets,dim=1).mean()
        loss_m2s = -torch.sum(F.log_softmax(sim_m2s,dim=1)*sim_m2s_targets,dim=1).mean()
        loss_s2im = -torch.sum(F.log_softmax(sim_s2im,dim=1)*sim_s2im_targets,dim=1).mean()
        loss_im2s = -torch.sum(F.log_softmax(sim_im2s,dim=1)*sim_im2s_targets,dim=1).mean()

        loss_sima = ((loss_s2i + loss_i2s)/2 + (loss_s2m + loss_m2s)/2 + (loss_s2im + loss_im2s)/2)/3

        self._dequeue_and_enqueue(selfies_feat_m,iupac_feat_m,mol_feat_m)


        # forward the postive selfies-iupac-mol pair
  

        iupac_mol_embeds = torch.cat([iupac_embeds,mol_embeds],dim=1)
        iupac_mol_mask = torch.cat([iupac_mask,mol_mask],dim=1)

        # mask = (batch_size,seq_len)

        # (batch_size,seq_len,embed_size)
        output_pos = self.fusion_encoder(selfies_embeds=selfies_embeds,
                                            iupac_mol_embeds=iupac_mol_embeds,
                                            selfies_mask=selfies_mask,
                                            iupac_mol_mask=iupac_mol_mask)

        with torch.no_grad():
            bs = selfies.size(0)
            weights_s2im = F.softmax(sim_s2im[:,:bs]+1e-4,dim=1)
            weights_im2s = F.softmax(sim_im2s[:,:bs]+1e-4,dim=1)

            weights_im2s.fill_diagonal_(0)
            weights_s2im.fill_diagonal_(0)

        # select a negative iupac_mol for each selfies
        # mask matrix of iupac_mol pair

        iupac_mol_embeds_neg = []
        iupac_mol_mask_neg = []
        for b in range(bs):
            neg_idx = torch.multinomial(weights_s2im[b],1).item()
            iupac_mol_embeds_neg.append(iupac_mol_embeds[neg_idx])
            iupac_mol_mask_neg.append(iupac_mol_mask[neg_idx])
        iupac_mol_embeds_neg = torch.stack(iupac_mol_embeds_neg,dim=0)
        iupac_mol_mask_neg = torch.stack(iupac_mol_mask_neg,dim=0)

        # select a negative selfies for each iupac_mol
        selfies_embeds_neg = []
        selfies_mask_neg = []
        for b in range(bs):
            neg_idx = torch.multinomial(weights_im2s[b],1).item()
            selfies_embeds_neg.append(selfies_embeds[neg_idx])
            selfies_mask_neg.append(selfies_mask[neg_idx])
        selfies_embeds_neg = torch.stack(selfies_embeds_neg,dim=0)
        selfies_mask_neg = torch.stack(selfies_mask_neg,dim=0)

        selfies_embeds_all = torch.cat([selfies_embeds,selfies_embeds_neg],dim=0)
        selfies_mask_all = torch.cat([selfies_mask,selfies_mask_neg],dim=0)

        iupac_mol_embeds_all = torch.cat([iupac_mol_embeds_neg,iupac_mol_embeds],dim=0)
        iuapc_mol_mask_all = torch.cat([iupac_mol_mask_neg,iupac_mol_mask],dim=0)  # to be modified


        # (2*batch_size,seq_len,hidden_size)
        output_neg = self.fusion_encoder(selfies_embeds=selfies_embeds_all,
                                        iupac_mol_embeds=iupac_mol_embeds_all,
                                        selfies_mask=selfies_mask_all,
                                        iupac_mol_mask=iuapc_mol_mask_all)
        
        sim_embeddings = torch.cat([output_pos[:,0,:],output_neg[:,0,:]],dim=0)
        sim_output = self.sim_head(sim_embeddings)
        simm_labels = torch.cat([torch.ones(bs,dtype=torch.long),torch.zeros(2*bs,dtype=torch.long)],dim=0).to(selfies.device)
        loss_simm = F.cross_entropy(sim_output,simm_labels)


    # ## mask selfies modeling loss ##
        input_ids = selfies.clone()
        labels = selfies.clone()

        pro_matrix = torch.full(labels.shape, self.msm_pro)
        input_ids, labels = self.mask(input_ids=input_ids,vocab_size=self.selfies_encoder.config.vocab_size,targets=labels,
                                      pro_matrix = pro_matrix,device=input_ids.device)  

         
        out,loss_msm = self.BertMaskSelfies(selfies_ids=input_ids,
                                                iupac_ids = iupac.clone(),
                                                mol_ids = mol.clone(),
                                                labels=labels,
                                                mode='train') 
            

        return out,loss_sima,loss_simm,loss_msm
        
    def get_fusion_feature(self,mol, iupac, selfies):
        with torch.no_grad():
            self.temp.clamp_(0.001,0.5)

      
        input_ids = selfies.clone()
        labels = selfies.clone()
        out = self.BertMaskSelfies(selfies_ids=input_ids,
                                                iupac_ids = iupac.clone(),
                                                mol_ids = mol.clone(),
                                                labels=None,
                                                mode='output') 
        return out
    def test(self,selfies,iupac,mol):
        input_ids = selfies.clone()
        labels = selfies.clone()

        pro_matrix = torch.full(labels.shape, self.msm_pro)
        input_ids, labels = self.mask(input_ids=input_ids,vocab_size=self.selfies_encoder.config.vocab_size,targets=labels,
                                      pro_matrix = pro_matrix,device=input_ids.device)  
        correct,num = self.BertMaskSelfies(selfies_ids=input_ids,
                                            iupac_ids = iupac.clone(),
                                            mol_ids = mol.clone(),
                                            labels=labels,
                                            mode='test') 
        return correct,num



    def mask(self,input_ids,vocab_size,targets=None,pro_matrix=None,device=None):
        masked_indices = torch.bernoulli(pro_matrix).bool()
                                               
        masked_indices[input_ids == 0] = False
        masked_indices[input_ids == 1] = False
        
        if targets is not None:
            targets[~masked_indices] = 0          

        # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
        indices_replaced = torch.bernoulli(torch.full(input_ids.shape, 0.8)).bool() & masked_indices
        input_ids[indices_replaced] = 2

        # 10% of the time, we replace masked input tokens with random word
        indices_random = torch.bernoulli(torch.full(input_ids.shape, 0.5)).bool() & masked_indices & ~indices_replaced
        random_words = torch.randint(vocab_size, input_ids.shape, dtype=torch.long).to(device)
        input_ids[indices_random] = random_words[indices_random]                     
        # The rest of the time (10% of the time) we keep the masked input tokens unchanged   
        
        if targets is not None:
            return input_ids, targets
        else:
            return input_ids



    @torch.no_grad() 
    def copy_params(self):
        for model_pair in self.model_pairs:           
            for param, param_m in zip(model_pair[0].parameters(), model_pair[1].parameters()):
                param_m.data.copy_(param.data) 
                param_m.requires_grad = False    

    @torch.no_grad()
    def _momentum_update(self):
        for model_pair in self.model_pairs:           
            for param, param_m in zip(model_pair[0].parameters(), model_pair[1].parameters()):
                param_m.data = param_m.data * self.momentum + param.data * (1. - self.momentum)

    @torch.no_grad()
    def _dequeue_and_enqueue(self,selfies_feat,iupac_feat,mol_feat):
        # selfies_feat = concat_all_gather(selfies_feat)
        # iupac_feat = concat_all_gather(iupac_feat)
        # mol_feat = concat_all_gather(mol_feat)

        batch_size = selfies_feat.shape[0]

        ptr = int(self.queue_ptr)
        assert self.quene_size % batch_size == 0  # for simplicity

        self.selfies_queue[:, ptr:ptr + batch_size] = selfies_feat.T
        self.iupac_queue[:, ptr:ptr + batch_size] = iupac_feat.T
        self.mol_queue[:, ptr:ptr + batch_size] = mol_feat.T

        ptr = (ptr + batch_size) % self.quene_size  # move pointer

        self.queue_ptr[0] = ptr



# @torch.no_grad()
# def concat_all_gather(tensor):
#     """
#     Performs all_gather operation on the provided tensors.
#     *** Warning ***: torch.distributed.all_gather has no gradient.
#     """
#     tensors_gather = [torch.ones_like(tensor)
#         for _ in range(torch.distributed.get_world_size())]
#     torch.distributed.all_gather(tensors_gather, tensor, async_op=False)

#     output = torch.cat(tensors_gather, dim=0)
#     return output


from craft_mol.batch_feature import BatchFeature
import os

model_path = './craft_mol/1m_checkpoint_09.pth'
data_path = '/home/hukun/code/Tmm-llama/data/molecular_description_generation/filtered_train.json'
output_dir = '/home/hukun/code/Tmm-llama/data/molecular_description_generation/train_mol_features'
if not os.path.exists(output_dir):
    os.makedirs(output_dir)
batch_size = 32
config_path = './craft_mol/inference.yaml'
batch_feature = BatchFeature(model_path=model_path, config_path=config_path)
# smiles = "C[C@@H1]1CN(S(=O)(=O)C2=C(C=C(C=C2)C#C[C@H1](C)O)O[C@H1]1CN(C)C(=O)NC(C)C)[C@@H1](C)CO"
# iupac_name = "1-[[(4R,5R)-8-[(3S)-3-hydroxybut-1-ynyl]-2-[(2S)-1-hydroxypropan-2-yl]-4-methyl-1,1-dioxo-4,5-dihydro-3H-6,1lambda6,2-benzoxathiazocin-5-yl]methyl]-1-methyl-3-propan-2-ylurea"
# mol_feature = batch_feature.get_feature_single(smiles=smiles, iupac_name=iupac_name)
# print(mol_feature.shape)
batch_feature.get_feature(data_path, batch_size, output_dir)



from model.TMMF import TMMF
from model.bert import BertForFinetune
from utils.utils import MetricLogger, SmoothedValue,AttrDict
from dataset import create_dataset, create_dataloader
from scheduler import create_scheduler
from optim import create_optimizer
import os
import ruamel.yaml as yaml
import numpy as np
import random
import time
import datetime
import json
from pathlib import Path
import torch
class Config(dict):
    def __init__(self, *args, **kwargs):
        super(Config, self).__init__(*args, **kwargs)

    def __getattr__(self, key):
        value = self[key]
        if isinstance(value, dict):
            value = Config(value)
        return value
    

def train(model, data_loader, optimizer, epoch, warmup_steps, device, scheduler, config):

    model.train()
    metric_logger = MetricLogger(delimiter="  ")
    metric_logger.add_meter('lr', SmoothedValue(window_size=50, fmt='{value:.6f}'))
    metric_logger.add_meter('loss', SmoothedValue(window_size=50, fmt='{value:.4f}'))

    
    header = 'Train Epoch: [{}]'.format(epoch)
    print_freq = 5   
    step_size = 100
    warmup_iterations = warmup_steps*step_size  


    for i, batch in enumerate(metric_logger.log_every(data_loader, print_freq, header)):
        
        optimizer.zero_grad()
        selfies,iupac,mol,label = batch['sfs_tokens']['selfies_tokens'],batch['ipc_tokens']['iupac_tokens'],batch['mlg'],batch['label']
        selfies = selfies.to(device)
        iupac = iupac.to(device)
        mol = mol.to(device)  
        label = label.to(device)

        _,loss = model._train(selfies_ids=selfies,iupac_ids=iupac,mol_ids=mol,labels=label)
        loss.backward()
        optimizer.step()    
        
        metric_logger.update(loss=loss.item())
        metric_logger.update(lr=optimizer.param_groups[0]["lr"])         
        
        if epoch==0 and i%step_size==0 and i<=warmup_iterations: 
            scheduler.step(i//step_size)        


    print("Averaged stats:", metric_logger.global_avg()) 
    return {k: "{:.3f}".format(meter.global_avg) for k, meter in metric_logger.meters.items()}

def val(model, data_loader,epoch,device):

    print_freq = 5   
    model.eval()
    loss_list = []
    with torch.no_grad():
        for i,batch in enumerate((data_loader)):
            selfies,iupac,mol,label = batch['sfs_tokens']['selfies_tokens'],batch['ipc_tokens']['iupac_tokens'],batch['mlg'],batch['label']
            selfies = selfies.to(device)
            iupac = iupac.to(device)
            mol = mol.to(device)  
            label = label.to(device)

            _,loss = model._train(selfies_ids=selfies,iupac_ids=iupac,mol_ids=mol,labels=label)

            loss_list.append(loss.item())
            if (i+1)%print_freq == 0:
                print(f"Epoch [{i}/{len(data_loader)}]] loss: {loss.item()}")
    print("val loss:",sum(loss_list)/len(loss_list))

from sklearn.metrics import roc_auc_score
def calculate_auc(model,loader,device):
    import math
    model.eval()
    y_true = []
    y_pred = []
    with torch.no_grad():
        for batch in loader:
            selfies,iupac,mol,label = batch['sfs_tokens']['selfies_tokens'],batch['ipc_tokens']['iupac_tokens'],batch['mlg'],batch['label']
            selfies = selfies.to(device)
            iupac = iupac.to(device)
            mol = mol.to(device)  
            label = label.to(device)
            output,labels = model._test(selfies,iupac,mol,label)
            y_true.extend(labels.cpu().numpy())
            y_pred.extend(output.cpu().numpy())
        # print("accu:",total_correct/total_num)
        y_true = np.array(y_true)
        y_pred = np.array(y_pred)
        roc_list = []
        for i in range(y_true.shape[1]):
            #AUC is only defined when there is at least one positive data.
            if np.sum(y_true[:,i] == 1) > 0 and np.sum(y_true[:,i] == -1) > 0:
                is_valid = y_true[:,i]**2 > 0
                roc_list.append(roc_auc_score((y_true[is_valid,i] + 1)/2, y_pred[is_valid,i]))

        if len(roc_list) < y_true.shape[1]:
            print("Some target is missing!")
            print("Missing ratio: %f" %(1 - float(len(roc_list))/y_true.shape[1]))
    return sum(roc_list)/len(roc_list),y_pred,y_true



def calculate_rsquared(model,loader,device):
    import math
    model.eval()
    y_true = []
    y_pred = []

    with torch.no_grad():
        for batch in loader:
            selfies,iupac,mol,label = batch['sfs_tokens']['selfies_tokens'],batch['ipc_tokens']['iupac_tokens'],batch['mlg'],batch['label']
            selfies = selfies.to(device)
            iupac = iupac.to(device)
            mol = mol.to(device)  
            label = label.to(device)
            output,labels = model._test(selfies,iupac,mol,label)


            
            y_true.extend(labels.cpu().numpy())
            y_pred.extend(output.cpu().numpy())

    return y_pred,y_true


def main(name):
    config = yaml.load(open('./config/finetune.yaml', 'r'), Loader=yaml.Loader)
    device = torch.device("cuda:0")
    seed = [42,52,62,72,32,22,12,2]
    max_epoch = config['schedular']['epochs']
    warmup_steps = config['schedular']['warmup_epochs']   
    print(f"Finetune on {name}......")
    print("creating dataset")
    dataset_config = Config(json.load(open(config['dataset'][name])))
    roc_list = []
    rmse_list = []
    r2_list = []
    for i in range(len(seed)):
        train_dataset,val_dataset,test_dataset = create_dataset(
        dataset_config['dataset']['ids_path'],
        dataset_config['dataset']['ipc_path'],
        dataset_config['dataset']['sfs_path'],
        dataset_config['dataset']['mlg_path'],
        dataset_config['dataset']['frac'],mode="finetune")


        print("creating model")
        pretrain_config = yaml.load(open('./config/pre_train.yaml', 'r'), Loader=yaml.Loader)
        model = TMMF(pretrain_config)
        model = model.to(device)

        path = './pre_train/1m_checkpoint_09.pth'
        checkpoint = torch.load(path)
        model.load_state_dict(checkpoint['model'])


        finetune_model = BertForFinetune(dataset_config,model.selfies_encoder,model.iupac_encoder,model.mol_encoder,model.fusion_encoder)
        finetune_model = finetune_model.to(device)
        Total_params = 0
        Trainable_params = 0
        NonTrainable_params = 0
        for param in finetune_model.parameters():
            mulValue = np.prod(param.size())  # 
            Total_params += mulValue  
            if param.requires_grad:
                Trainable_params += mulValue  # trainable
            else:
                NonTrainable_params += mulValue  


        print(f'Total params: {Total_params}')
        print(f'Trainable params: {Trainable_params}')
        print(f'Non-trainable params: {NonTrainable_params}')
        torch.manual_seed(seed[i])
        np.random.seed(seed[i])
        train_loader = create_dataloader(train_dataset,dataset_config['batch_size'],mode="finetune",task_type=dataset_config['dataset']['task_type'])
        val_loader = create_dataloader(val_dataset,dataset_config['batch_size'],mode="finetune",task_type=dataset_config['dataset']['task_type'])
        test_loader = create_dataloader(test_dataset,dataset_config['batch_size'],mode="finetune",task_type=dataset_config['dataset']['task_type'])


        arg_opt = AttrDict(config['optimizer'])
        optimizer = create_optimizer(arg_opt, finetune_model)
        arg_sche = AttrDict(config['schedular'])
        lr_scheduler, _ = create_scheduler(arg_sche, optimizer)  

        print("Start training")
        start_time = time.time()
        start_epoch = 0

        for epoch in range(start_epoch, max_epoch):
            
            if epoch>0:
                lr_scheduler.step(epoch+warmup_steps)  
            
            train_stats = train(finetune_model, train_loader, optimizer, epoch, warmup_steps, device, lr_scheduler, config)
            
            log_stats = {**{f'train_{k}': v for k, v in train_stats.items()},
                            'epoch': epoch}                     


            save_path = 'downstream/' + dataset_config.dataset.name
            if not os.path.exists(save_path):
                os.mkdir(save_path)
            with open(os.path.join(save_path + "/log"+ str(i) + ".txt"),"a") as f:
                f.write(json.dumps(log_stats) + "\n")
            
            print("Start validation")
            val(finetune_model,val_loader,epoch,device)

        

        total_time = time.time() - start_time
        total_time_str = str(datetime.timedelta(seconds=int(total_time)))
        print('Training time {}'.format(total_time_str)) 

        print("Start testing")
        if dataset_config.dataset.task_type == "classification":
            roc_auc_score,y_pred,y_true = calculate_auc(finetune_model,test_loader,device)
            save_name = 'roc'
            roc_list.append(roc_auc_score)
            print(roc_auc_score)
            

        elif dataset_config.dataset.task_type == "regression":
            y_pred,y_true = calculate_rsquared(finetune_model,test_loader,device)
            from sklearn.metrics import r2_score,mean_squared_error
            print("r2_score",r2_score(y_true,y_pred))
            print("RMSE",np.sqrt(mean_squared_error(y_true,y_pred,squared=False)))
            save_name = 'r2'
            rmse_list.append(np.sqrt(mean_squared_error(y_true,y_pred,squared=False)))
            r2_list.append(r2_score(y_true,y_pred))
        
    if dataset_config.dataset.task_type == "classification":
        eval_data = {'seed':seed,'roc':roc_list}
    elif dataset_config.dataset.task_type == "regression":
        eval_data = {'seed':seed,'rmse':rmse_list,'r2':r2_list}
    from pandas.core.frame import DataFrame
    eval_data = DataFrame(eval_data)
    eval_data.to_csv(save_path + '/' + save_name+'.csv')

    save_obj = {'model': model.state_dict()}
    torch.save(save_obj, os.path.join('./pre_train/', f'{name}_checkpoint.pth'))



if __name__ == '__main__':
    dataset_name = ['bbbp','bace' ,'clintox']
    dataset_name = ['bbbp','bace','clintox','esol','freesolv' ,'lipo']
    dataset_name = ['bace','clintox','tox21','sider','esol','freesolv']
    dataset_name = ['lipo']
    for name in dataset_name:
        main(name)